"""Starrail chronicle challenge."""

import typing

import pydantic

from genshin.models.model import Aliased, APIModel
from genshin.models.starrail.character import FloorCharacter

from .base import PartialTime

__all__ = (
    "APCShadowBoss",
    "APCShadowFloor",
    "APCShadowFloorNode",
    "APCShadowSeason",
    "AnomalyArbitration",
    "AnomalyBossInfo",
    "AnomalyBossRecord",
    "AnomalyMiniBossInfo",
    "AnomalyMiniBossRecord",
    "AnomalyPlayer",
    "AnomalyRecord",
    "AnomalySeason",
    "AnomalySummary",
    "ChallengeBuff",
    "FictionFloor",
    "FictionFloorNode",
    "FloorNode",
    "StarRailAPCShadow",
    "StarRailChallenge",
    "StarRailChallengeFloor",
    "StarRailChallengeSeason",
    "StarRailFloor",
    "StarRailPureFiction",
)


class FloorNode(APIModel):
    """Node for a memory of chaos floor."""

    challenge_time: PartialTime
    avatars: list[FloorCharacter]


class StarRailChallengeFloor(APIModel):
    """Base model for star rail challenge floors."""

    id: int = Aliased("maze_id")
    name: str
    star_num: int
    is_quick_clear: bool = Aliased("is_fast")

    @pydantic.computed_field  # type: ignore[prop-decorator]
    @property
    def stars(self) -> int:
        """Number of stars earned on the floor."""
        return self.star_num


class StarRailFloor(StarRailChallengeFloor):
    """Floor in a memory of chaos challenge."""

    round_num: int
    is_chaos: bool
    node_1: FloorNode
    node_2: FloorNode


class StarRailChallengeSeason(APIModel):
    """A season of a challenge."""

    id: int = Aliased("schedule_id")
    name: str = Aliased("name_mi18n")
    status: str
    begin_time: PartialTime
    end_time: PartialTime


class StarRailChallenge(APIModel):
    """Memory of chaos challenge in a season."""

    name: str
    season: int = Aliased("schedule_id")
    begin_time: typing.Optional[PartialTime]
    end_time: typing.Optional[PartialTime]

    total_stars: int = Aliased("star_num")
    max_floor: str
    total_battles: int = Aliased("battle_num")
    has_data: bool

    floors: typing.Sequence[StarRailFloor] = Aliased("all_floor_detail")
    seasons: typing.Sequence[StarRailChallengeSeason] = Aliased("groups")

    @pydantic.model_validator(mode="before")
    def __extract_name(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if "groups" in values and isinstance(values["groups"], list):
            seasons = typing.cast(list[dict[str, typing.Any]], values["groups"])
            if len(seasons) > 0:
                values["name"] = seasons[0]["name_mi18n"]
            else:
                values["name"] = ""

        return values


class ChallengeBuff(APIModel):
    """Buff used in a pure fiction or apocalyptic shadow node."""

    id: int
    name: str = Aliased("name_mi18n")
    description: str = Aliased("desc_mi18n")
    icon: str


class FictionFloorNode(FloorNode):
    """Node for a Pure Fiction floor."""

    buff: typing.Optional[ChallengeBuff]
    score: int


class FictionFloor(StarRailChallengeFloor):
    """Floor in a Pure Fiction challenge."""

    round_num: int
    node_1: FictionFloorNode
    node_2: FictionFloorNode

    @property
    def score(self) -> int:
        """Total score of the floor."""
        return self.node_1.score + self.node_2.score


class StarRailPureFiction(APIModel):
    """Pure Fiction challenge in a season."""

    name: str = pydantic.Field(deprecated="Use `season_id` together with `seasons instead`.")
    season_id: int = pydantic.Field(deprecated="Use `season_id` together with `seasons instead`.")
    begin_time: PartialTime = pydantic.Field(deprecated="Use `season_id` together with `seasons instead`.")
    end_time: PartialTime = pydantic.Field(deprecated="Use `season_id` together with `seasons instead`.")

    total_stars: int = Aliased("star_num")
    max_floor: str
    total_battles: int = Aliased("battle_num")
    has_data: bool

    floors: list[FictionFloor] = Aliased("all_floor_detail")
    seasons: list[StarRailChallengeSeason] = Aliased("groups")
    max_floor_id: int

    @pydantic.model_validator(mode="before")
    def __unnest_groups(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if "groups" in values and isinstance(values["groups"], list):
            seasons = typing.cast(list[dict[str, typing.Any]], values["groups"])
            if len(seasons) > 0:
                values["name"] = seasons[0]["name_mi18n"]
                values["season_id"] = seasons[0]["schedule_id"]
                values["begin_time"] = seasons[0]["begin_time"]
                values["end_time"] = seasons[0]["end_time"]

        return values


class APCShadowFloorNode(FloorNode):
    """Node for a apocalyptic shadow floor."""

    challenge_time: typing.Optional[PartialTime]  # type: ignore[assignment]
    buff: typing.Optional[ChallengeBuff]
    score: int
    boss_defeated: bool

    @property
    def has_data(self) -> bool:
        """Check if the node has data."""
        return bool(self.avatars)


class APCShadowFloor(StarRailChallengeFloor):
    """Floor in an apocalyptic shadow challenge."""

    node_1: APCShadowFloorNode
    node_2: APCShadowFloorNode
    last_update_time: PartialTime
    is_quick_clear: bool = Aliased("is_fast")

    @property
    def score(self) -> int:
        """Total score of the floor."""
        return self.node_1.score + self.node_2.score


class APCShadowBoss(APIModel):
    """Boss in an apocalyptic shadow challenge."""

    id: int
    name_mi18n: str
    icon: str


class APCShadowSeason(StarRailChallengeSeason):
    """Season of an apocalyptic shadow challenge."""

    upper_boss: APCShadowBoss
    lower_boss: APCShadowBoss


class StarRailAPCShadow(APIModel):
    """Apocalyptic shadow challenge in a season."""

    total_stars: int = Aliased("star_num")
    max_floor: str
    total_battles: int = Aliased("battle_num")
    has_data: bool

    floors: list[APCShadowFloor] = Aliased("all_floor_detail")
    seasons: list[APCShadowSeason] = Aliased("groups")
    max_floor_id: int


class AnomalySeason(APIModel):
    """Anomaly Arbitration season."""

    id: int = Aliased("group_id")
    begin_time: PartialTime
    end_time: PartialTime
    status: str
    name: str = Aliased("name_mi18n")
    game_version: str
    icon: str = Aliased("theme_pic_path")


class AnomalyBossInfo(APIModel):
    """Anomaly Arbitration boss info."""

    id: int = Aliased("maze_id")
    name: str = Aliased("name_mi18n")
    game_mode_name: str = Aliased("hard_mode_name_mi18n")
    icon: str


class AnomalyMiniBossInfo(APIModel):
    """Anomaly Arbitration mini boss info."""

    id: int = Aliased("maze_id")
    level_name: str = Aliased("name")
    name: str = Aliased("monster_name")
    icon: str = Aliased("monster_icon")


class AnomalyBossRecord(APIModel):
    """Anomaly Arbitration boss record."""

    id: int = Aliased("maze_id")
    has_data: bool = Aliased("has_challenge_record")
    time: PartialTime = Aliased("challenge_time")
    characters: list[FloorCharacter] = Aliased("avatars")
    buff: ChallengeBuff
    is_hard_mode: bool = Aliased("hard_mode")
    cycles_used: int = Aliased("round_num")
    stars: int = Aliased("star_num")
    has_color_medal: bool = Aliased("finish_color_medal")
    medal_type: str = Aliased("challenge_peak_rank_icon_type")
    medal_icon: str = Aliased("challenge_peak_rank_icon")
    record_unique_key: str


class AnomalyMiniBossRecord(APIModel):
    """Anomaly Arbitration mini boss record."""

    id: int = Aliased("maze_id")
    has_data: bool = Aliased("has_challenge_record")
    time: typing.Optional[PartialTime] = Aliased("challenge_time")
    characters: list[FloorCharacter] = Aliased("avatars")
    cycles_used: int = Aliased("round_num")
    stars: int = Aliased("star_num")
    is_quick_clear: bool = Aliased("is_fast")


class AnomalyRecord(APIModel):
    """HSR Anomaly Arbitration record."""

    season: AnomalySeason = Aliased("group")
    boss: AnomalyBossInfo = Aliased("boss_info")
    mini_bosses: list[AnomalyMiniBossInfo] = Aliased("mob_infos")
    has_data: bool = Aliased("has_challenge_record")
    cycles_used: int = Aliased("battle_num")

    boss_record: typing.Optional[AnomalyBossRecord] = Aliased("boss_record", default=None)
    mini_boss_records: list[AnomalyMiniBossRecord] = Aliased("mob_records")

    boss_stars: int
    mini_boss_stars: int = Aliased("mob_stars")


class AnomalySummary(APIModel):
    """HSR Anomaly Arbitration summary."""

    challenge_attempts: int = Aliased("total_battle_num")
    mini_boss_stars: int = Aliased("mob_stars")
    boss_stars: int
    medal_type: str = Aliased("challenge_peak_rank_icon_type")
    medal_icon: str = Aliased("challenge_peak_rank_icon")


class AnomalyPlayer(APIModel):
    """HSR Anomaly Arbitration player info."""

    server: str
    nickname: str
    level: int
    uid: str = Aliased("role_id")


class AnomalyArbitration(APIModel):
    """HSR Anomaly Arbitration info."""

    records: list[AnomalyRecord] = Aliased("challenge_peak_records")
    summary: AnomalySummary = Aliased("challenge_peak_best_record_brief")
    player: AnomalyPlayer = Aliased("role")



"""Genshin models."""

from .calculator import *
from .character import *
from .chronicle import *
from .constants import *
from .daily import *
from .diary import *
from .gacha import *
from .lineup import *
from .teapot import *
from .transaction import *
from .wiki import *



"""Transaction client."""

import functools
import typing
import urllib.parse

from genshin import paginators, utility
from genshin.client import routes
from genshin.client.components import base
from genshin.models.genshin import transaction as models

__all__ = ["TransactionClient"]


class TransactionClient(base.BaseClient):
    """Transaction component."""

    async def request_transaction(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the transaction log endpoint."""
        params = dict(params or {})
        authkey = authkey or self.authkey

        if authkey is None:
            raise RuntimeError("No authkey provided")

        base_url = routes.YSULOG_URL.get_url(self.region)
        url = base_url / endpoint

        params["authkey_ver"] = 1
        params["sign_type"] = 2
        params["authkey"] = urllib.parse.unquote(authkey)
        params["lang"] = utility.create_short_lang_code(lang or self.lang)

        return await self.request(url, method=method, params=params, **kwargs)

    async def _get_transaction_page(
        self,
        end_id: int,
        kind: str,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.BaseTransaction]:
        """Get a single page of transactions."""
        kind = models.TransactionKind(kind)
        endpoint = "Get" + kind.value.capitalize() + "Log"

        data = await self.request_transaction(
            endpoint,
            lang=lang,
            authkey=authkey,
            params=dict(end_id=end_id, size=20),
        )

        transactions: list[models.BaseTransaction] = []
        for trans in data["list"]:
            model = models.ItemTransaction if "name" in trans else models.Transaction
            model = typing.cast("type[models.BaseTransaction]", model)
            transactions.append(model(**trans, kind=kind))

        return transactions

    def transaction_log(
        self,
        kind: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.BaseTransaction]:
        """Get the transaction log of a user."""
        kinds = kind or ["primogem", "crystal", "resin", "artifact", "weapon"]

        if isinstance(kinds, str):
            kinds = [kinds]

        iterators: list[paginators.Paginator[models.BaseTransaction]] = []
        for kind in kinds:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_transaction_page,
                        kind=kind,
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda trans: trans.time.timestamp())



"""Deprecation decorator."""

import functools
import inspect
import typing
import warnings

__all__ = ["deprecated", "warn_deprecated"]

CallableT = typing.TypeVar("CallableT", bound=typing.Callable[..., typing.Any])


def warn_deprecated(
    obj: typing.Any,
    *,
    alternative: typing.Optional[str] = None,
    stack_level: int = 3,
) -> None:
    """Raise a deprecated warning."""
    if inspect.isclass(obj) or inspect.isfunction(obj):
        obj = f"{obj.__qualname__}"

    message = f"{obj} is deprecated and will be removed in the following version."

    if alternative is not None:
        message += f" You can use '{alternative}' instead."

    warnings.warn(message, category=DeprecationWarning, stacklevel=stack_level)


def deprecated(alternative: typing.Optional[str] = None) -> typing.Callable[[CallableT], CallableT]:
    """Mark a function as deprecated."""

    def decorator(obj: CallableT) -> CallableT:
        alternative_str = f"You can use `{alternative}` instead." if alternative else ""

        doc = inspect.getdoc(obj) or ""
        doc += (
            "\n\n"
            "!!! warning\n"
            f"    This function is deprecated and will be removed in the following version.\n"
            f"    {alternative_str}\n"
        )
        obj.__doc__ = doc

        @functools.wraps(obj)
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            warn_deprecated(obj, alternative=alternative, stack_level=3)
            return obj(*args, **kwargs)

        return typing.cast("CallableT", wrapper)

    return decorator



"""Teapot component."""

import functools
import typing

from genshin import paginators, utility
from genshin.client import routes
from genshin.client.components import base
from genshin.models.genshin import teapot as models

__all__ = ["TeapotClient"]


class TeapotClient(base.BaseClient):
    """teapot component."""

    async def request_teapot(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the teapot endpoint."""
        params = dict(params or {})

        base_url = routes.TEAPOT_URL.get_url(self.region)
        url = base_url / endpoint

        params["lang"] = lang or self.lang

        return await self.request(url, method=method, params=params, **kwargs)

    async def _get_teapot_replica_page(
        self,
        page: int,
        *,
        zip_type: int = 1,
        block_id: typing.Optional[str] = None,
        module_id: typing.Optional[str] = None,
        region: typing.Optional[str] = None,
        version: typing.Optional[str] = None,
        limit: int = 20,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.TeapotReplica]:
        """Get a teapot replica page."""
        params = dict(
            next=page * limit,  # weirdest sorting ever
            zip_type=zip_type,
            block_id=block_id or "",
            module_id=module_id or "",
            target_region=region or "",
            version=version or "",
            limit=limit,
        )
        data = await self.request_teapot("list", lang=lang, params=params)
        return [models.TeapotReplica(**entry) for entry in data["articles"]]

    def teapot_replicas(
        self,
        *,
        limit: typing.Optional[int] = None,
        zip_type: int = 1,
        block_id: typing.Optional[str] = None,
        module_id: typing.Optional[str] = None,
        region: typing.Optional[str] = None,
        uid: typing.Optional[int] = None,
        version: typing.Optional[str] = None,
        page_size: int = 20,
        lang: typing.Optional[str] = None,
    ) -> paginators.PagedPaginator[models.TeapotReplica]:
        """Get a teapot replica paginator."""
        if not region and uid:
            region = utility.recognize_genshin_server(uid)

        return paginators.PagedPaginator(
            functools.partial(
                self._get_teapot_replica_page,
                zip_type=zip_type,
                block_id=block_id,
                module_id=module_id,
                region=region,
                version=version,
                limit=page_size,
                lang=lang,
            ),
            limit=limit,
            page_size=page_size,
        )



from typing import Any, Sequence

from pydantic import field_validator

from genshin import types
from genshin.models.model import Aliased, APIModel

__all__ = (
    "AccompanyCharacter",
    "AccompanyCharacterAttribute",
    "AccompanyCharacterGame",
    "AccompanyCharacterInfo",
    "AccompanyCharacterProfile",
    "AccompanyResult",
)


class AccompanyCharacterInfo(APIModel):
    """Accompany character info."""

    role_id: int
    game_id: int
    name: str
    brief_name: str
    topic_id: int
    game_name: str


class AccompanyCharacterProfile(APIModel):
    """Accompany character color and images."""

    card_color: str
    bg_dark_color: str
    bg_light_color: str
    bg_card_color: str

    card_image: str = Aliased("card_img_url")
    video: str = Aliased("full_screen_video_265")
    icon: str = Aliased("icon_url")


class AccompanyCharacter(APIModel):
    """Accompany character."""

    info: AccompanyCharacterInfo = Aliased("basic")
    profile: AccompanyCharacterProfile = Aliased("attr_profile")
    attribute_ids: Sequence[int]


class AccompanyCharacterAttribute(APIModel):
    """Accompany character attribute."""

    id: int
    icon: str = Aliased("icon_url")
    corner_icon: str = Aliased("corner_icon_url")


class AccompanyCharacterGame(APIModel):
    """Accompany character game."""

    attributes: Sequence[Sequence[AccompanyCharacterAttribute]] = Aliased("attribute_group_list")
    id: int = Aliased("game_id")
    icon: str = Aliased("game_icon")
    characters: Sequence[AccompanyCharacter] = Aliased("role_list")

    @field_validator("attributes", mode="before")
    def __unnest_attributes(cls, v: list[dict[str, Any]]) -> list[list[dict[str, Any]]]:
        return [group["attribute_list"] for group in v]

    @property
    def game(self) -> types.Game:
        if self.id == 2:
            return types.Game.GENSHIN
        if self.id == 6:
            return types.Game.STARRAIL
        if self.id == 8:
            return types.Game.ZZZ

        raise ValueError(f"Unknown game ID: {self.id}")


class AccompanyResult(APIModel):
    """Accompany result."""

    accompany_days: int
    points_increased: int = Aliased("increase_accompany_point")



"""Ratelimit handlers."""

import logging
import typing

import aiohttp
from tenacity import before_sleep_log, retry, retry_if_exception_type, stop_after_attempt, wait_random_exponential

from genshin import errors

LOGGER_ = logging.getLogger(__name__)
TIMEOUT_ERRORS = (TimeoutError, aiohttp.ClientError, ConnectionResetError)
CallableT = typing.TypeVar("CallableT", bound=typing.Callable[..., typing.Awaitable[typing.Any]])


def handle_ratelimits(
    tries: int = 10,
    exception: type[errors.GenshinException] = errors.VisitsTooFrequently,
    delay: float = 0.5,
) -> typing.Callable[[CallableT], CallableT]:
    """Handle ratelimits for requests."""
    return retry(
        stop=stop_after_attempt(tries),
        wait=wait_random_exponential(multiplier=delay, min=delay),
        retry=retry_if_exception_type(exception),
        reraise=True,
        before_sleep=before_sleep_log(LOGGER_, logging.DEBUG),
    )


def handle_request_timeouts(
    tries: int = 10,
    delay: float = 0.5,
) -> typing.Callable[[CallableT], CallableT]:
    """Handle timeout errors for requests."""
    try:
        from aiohttp_socks import ProxyError
    except ImportError:
        return retry(
            stop=stop_after_attempt(tries),
            wait=wait_random_exponential(multiplier=delay, min=delay),
            retry=retry_if_exception_type(TIMEOUT_ERRORS),
            reraise=True,
            before_sleep=before_sleep_log(LOGGER_, logging.DEBUG),
        )
    else:
        return retry(
            stop=stop_after_attempt(tries),
            wait=wait_random_exponential(multiplier=delay, min=delay),
            retry=retry_if_exception_type((ProxyError, *TIMEOUT_ERRORS)),
            reraise=True,
            before_sleep=before_sleep_log(LOGGER_, logging.DEBUG),
        )



"""Hoyolab component."""

import asyncio
import json
import time
import typing
import uuid
import warnings

import yarl

from genshin import constants, types, utility
from genshin.client import cache as client_cache
from genshin.client import routes
from genshin.client.components import base
from genshin.client.manager import managers
from genshin.models import hoyolab as models

__all__ = ["HoyolabClient"]


class HoyolabClient(base.BaseClient):
    """Hoyolab component."""

    async def _get_server_region(self, uid: int, game: types.Game) -> str:
        """Fetch the server region of an account from the API."""
        data = await self.request(
            routes.GET_USER_REGION_URL.get_url(),
            params=dict(game_biz=utility.get_prod_game_biz(self.region, game)),
            cache=client_cache.cache_key("server_region", game=game, uid=uid, region=self.region),
        )
        for account in data["list"]:
            if account["game_uid"] == str(uid):
                return account["region"]

        raise ValueError(f"Failed to recognize server for game {game!r} and uid {uid!r}")

    async def _request_announcements(
        self,
        game: types.Game,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Announcement]:
        """Get a list of game announcements."""
        if game is types.Game.GENSHIN:
            params = dict(
                game="hk4e",
                game_biz="hk4e_global",
                bundle_id="hk4e_global",
                platform="pc",
                region=utility.recognize_genshin_server(uid),
                uid=uid,
                level=8,
                lang=lang or self.lang,
            )
            url = routes.HK4E_URL.get_url()
        elif game is types.Game.ZZZ:
            params = dict(
                game="nap",
                game_biz="nap_global",
                bundle_id="nap_global",
                platform="pc",
                region=utility.recognize_zzz_server(uid),
                level=60,
                lang=lang or self.lang,
                uid=uid,
            )
            url = routes.NAP_URL.get_url()
        elif game is types.Game.STARRAIL:
            params = dict(
                game="hkrpg",
                game_biz="hkrpg_global",
                bundle_id="hkrpg_global",
                platform="pc",
                region=utility.recognize_starrail_server(uid),
                uid=uid,
                level=70,
                lang=lang or self.lang,
                channel_id=1,
            )
            url = routes.HKRPG_URL.get_url()
        else:
            msg = f"{game!r} is not supported yet."
            raise ValueError(msg)

        info, details = await asyncio.gather(
            self.request_hoyolab(
                url / "announcement/api/getAnnList",
                lang=lang,
                params=params,
            ),
            self.request_hoyolab(
                url / "announcement/api/getAnnContent",
                lang=lang,
                params=params,
            ),
        )

        announcements: list[typing.Dict[str, typing.Any]] = []
        extra_list: list[typing.Dict[str, typing.Any]] = (
            info["pic_list"][0]["type_list"] if "pic_list" in info and info["pic_list"] else []
        )

        for sublist in info["list"] + extra_list:
            for ann in sublist["list"]:
                detail = next((i for i in details["list"] if i["ann_id"] == ann["ann_id"]), None)

                # Update existing announcements with new details
                same_title = next((a for a in announcements if a["title"] == ann["title"]), None)
                if same_title is not None:
                    if ann.get("banner"):
                        same_title["banner"] = ann["banner"]
                    if ann.get("img"):
                        same_title["img"] = ann["img"]
                    continue

                announcements.append({**ann, **(detail or {})})

        return [models.Announcement(**i) for i in announcements]

    async def _request_mimo(
        self,
        endpoint: str,
        *,
        method: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
    ) -> typing.Any:
        game_id = params.get("game_id") if params else data.get("game_id")
        if game_id is None and self.game is None:
            raise ValueError("Cannot determine game for this traveling mimo request.")

        if game_id == 2 or self.game is types.Game.GENSHIN:
            url = routes.MIMO_URL.get_url() / "qiuqiu" / endpoint.replace("-", "_")
        else:
            url = routes.MIMO_URL.get_url() / endpoint
        return await self.request(url, method=method, params=params, data=data)

    async def search_users(
        self,
        keyword: str,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.PartialHoyolabUser]:
        """Search hoyolab users."""
        data = await self.request_bbs(
            "community/search/wapi/search/user",
            lang=lang,
            params=dict(keyword=keyword, page_size=20),
            cache=client_cache.cache_key("search", keyword=keyword, lang=self.lang),
        )
        return [models.PartialHoyolabUser(**i["user"]) for i in data["list"]]

    async def get_hoyolab_user(
        self,
        hoyolab_id: typing.Optional[int] = None,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.FullHoyolabUser:
        """Get a hoyolab user."""
        if self.region == types.Region.OVERSEAS:
            url = "/community/painter/wapi/user/full"
        elif self.region == types.Region.CHINESE:
            url = "/user/wapi/getUserFullInfo"
        else:
            raise TypeError(f"{self.region!r} is not a valid region.")

        data = await self.request_bbs(
            url=url,
            lang=lang,
            params=dict(uid=hoyolab_id) if hoyolab_id else None,
            cache=client_cache.cache_key("hoyolab", uid=hoyolab_id, lang=lang or self.lang),
        )
        return models.FullHoyolabUser(**data["user_info"])

    async def get_recommended_users(self, *, limit: int = 200) -> typing.Sequence[models.PartialHoyolabUser]:
        """Get a list of recommended active users."""
        warnings.warn("This endpoint is removed and an empty list will always be returned", DeprecationWarning)
        return []

    async def get_genshin_announcements(
        self,
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Announcement]:
        """Get a list of Genshin Impact announcements."""
        if self.cookie_manager.multi:
            uid = uid or await self._get_uid(types.Game.GENSHIN)
        else:
            uid = uid or 900000005
        return await self._request_announcements(types.Game.GENSHIN, uid, lang=lang)

    async def get_zzz_announcements(
        self,
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Announcement]:
        """Get a list of Zenless Zone Zero announcements."""
        if self.cookie_manager.multi:
            uid = uid or await self._get_uid(types.Game.ZZZ)
        else:
            uid = uid or 1300000000
        return await self._request_announcements(types.Game.ZZZ, uid, lang=lang)

    async def get_starrail_announcements(
        self,
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Announcement]:
        """Get a list of Star Rail announcements."""
        if self.cookie_manager.multi:
            uid = uid or await self._get_uid(types.Game.STARRAIL)
        else:
            uid = uid or 809162009
        return await self._request_announcements(types.Game.STARRAIL, uid, lang=lang)

    @managers.requires_cookie_token
    async def redeem_code(
        self,
        code: str,
        uid: typing.Optional[int] = None,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
        region: typing.Optional[str] = None,
    ) -> None:
        """Redeems a gift code for the current user."""
        if game is None:
            if self.default_game is None:
                raise RuntimeError("No default game set.")

            game = self.default_game

        if game not in constants.CODE_SUPPORTED_GAMES:
            raise ValueError(f"{game} does not support code redemption.")

        uid = uid or await self._get_uid(game)

        try:
            region = region or utility.recognize_server(uid, game)
        except Exception:
            warnings.warn(f"Failed to recognize server for game {game!r} and uid {uid!r}, fetching from API now.")
            region = await self._get_server_region(uid, game)

        post_request_games = {types.Game.ZZZ, types.Game.STARRAIL}
        game_origins = {
            types.Game.GENSHIN: "https://genshin.hoyoverse.com",
            types.Game.ZZZ: "https://zenless.hoyoverse.com",
            types.Game.STARRAIL: "https://hsr.hoyoverse.com",
            types.Game.TOT: "https://tot.hoyoverse.com",
        }
        headers = {
            "Origin": game_origins[game],
            "Referer": f"{game_origins[game]}/",
        }

        if game in post_request_games:
            data = dict(
                cdkey=code,
                device_uuid=str(uuid.uuid4()),
                game_biz=utility.get_prod_game_biz(self.region, game),
                lang=utility.create_short_lang_code(lang or self.lang),
                platform="4",
                region=region,
                t=int(time.time()),
                uid=uid,
            )
            await self.request(
                routes.CODE_URL.get_url(self.region, game),
                data=data,
                method="POST",
                headers=headers,
            )
        else:
            await self.request(
                routes.CODE_URL.get_url(self.region, game),
                params=dict(
                    uid=uid,
                    region=region,
                    lang=utility.create_short_lang_code(lang or self.lang),
                    cdkey=code,
                    game_biz=utility.get_prod_game_biz(self.region, game),
                ),
                method="GET",
                headers=headers,
            )

    @managers.no_multi
    async def check_in_community(self) -> None:
        """Check in to the hoyolab community and claim your daily 5 community exp."""
        raise RuntimeError("This API is deprecated.")

    @base.region_specific(types.Region.OVERSEAS)
    async def fetch_mi18n(
        self, url: typing.Union[str, yarl.URL], filename: str, *, lang: typing.Optional[str] = None
    ) -> typing.Mapping[str, str]:
        """Fetch a mi18n file."""
        return await self.request(
            yarl.URL(url) / f"{filename}/{filename}-{lang or self.lang}.json",
            cache=client_cache.cache_key("mi18n", filename=filename, url=url, lang=lang or self.lang),
        )

    @base.region_specific(types.Region.OVERSEAS)
    async def get_mimo_games(self, *, lang: typing.Optional[str] = None) -> typing.Sequence[models.MimoGame]:
        """Get a list of Traveling Mimo games."""
        data = await self._request_mimo("index", params=dict(lang=lang or self.lang))
        if self.game is None:
            raise RuntimeError("No default game set.")

        if self.game is types.Game.GENSHIN:
            return [models.MimoGame(**i["act_info"]) for i in data["act_list"]]
        return [models.MimoGame(**i) for i in data["list"]]

    @base.region_specific(types.Region.OVERSEAS)
    async def _get_mimo_game_data(
        self, game: typing.Union[typing.Literal["hoyolab"], types.Game]
    ) -> typing.Tuple[int, int]:
        games = await self.get_mimo_games()
        mimo_game = next((i for i in games if i.game == game), None)
        if mimo_game is None:
            raise ValueError(f"Game {game!r} not found in the list of Traveling Mimo games.")
        return mimo_game.id, mimo_game.version_id

    @base.region_specific(types.Region.OVERSEAS)
    async def _parse_mimo_args(
        self,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
    ) -> typing.Tuple[int, int]:
        if game_id is None or version_id is None:
            if game is None:
                if self.default_game is None:
                    raise RuntimeError("No default game set.")
                game = self.default_game

            if game not in {types.Game.GENSHIN, types.Game.ZZZ, types.Game.STARRAIL, "hoyolab"}:
                raise ValueError(f"{game!r} does not support Traveling Mimo.")
            game_id, version_id = await self._get_mimo_game_data(game)

        return game_id, version_id

    @base.region_specific(types.Region.OVERSEAS)
    async def get_mimo_tasks(
        self,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.MimoTask]:
        """Get a list of Traveling Mimo missions (tasks)."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        data = await self._request_mimo(
            "task-list",
            params=dict(game_id=game_id, lang=lang or self.lang, version_id=version_id),
        )
        return [models.MimoTask(**i) for i in data["task_list"]]

    @base.region_specific(types.Region.OVERSEAS)
    async def claim_mimo_task_reward(
        self,
        task_id: int,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> None:
        """Claim a Traveling Mimo mission (task) reward."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        await self._request_mimo(
            "receive-point",
            params=dict(task_id=task_id, game_id=game_id, lang=lang or self.lang, version_id=version_id),
            method="POST" if game_id == 2 else "GET",
        )

    @base.region_specific(types.Region.OVERSEAS)
    async def finish_mimo_task(
        self,
        task_id: int,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> None:
        """Finish a Traveling Mimo mission (task) reward."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        await self._request_mimo(
            "finish-task",
            data=dict(task_id=task_id, game_id=game_id, lang=lang or self.lang, version_id=version_id),
            method="POST",
        )

    @base.region_specific(types.Region.OVERSEAS)
    async def get_mimo_shop_items(
        self,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.MimoShopItem]:
        """Get a list of Traveling Mimo shop items."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        data = await self._request_mimo(
            "exchange-list",
            params=dict(game_id=game_id, lang=lang or self.lang, version_id=version_id),
        )
        return [models.MimoShopItem(**i) for i in data["exchange_award_list"]]

    @base.region_specific(types.Region.OVERSEAS)
    async def buy_mimo_shop_item(
        self,
        item_id: int,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> str:
        """Buy an item from the Traveling Mimo shop and return a gift code to redeem it."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        data = await self._request_mimo(
            "exchange",
            data=dict(award_id=item_id, game_id=game_id, lang=lang or self.lang, version_id=version_id),
            method="POST",
        )
        return data["exchange_code"]

    @base.region_specific(types.Region.OVERSEAS)
    async def get_mimo_point_count(
        self,
        *,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
    ) -> int:
        """Get the current Traveling Mimo point count."""
        game = game or self.default_game
        games = await self.get_mimo_games()
        mimo_game = next((i for i in games if i.game == game), None)
        if mimo_game is None:
            raise ValueError(f"Game {game!r} not found in the list of Traveling Mimo games.")
        return mimo_game.point

    @base.region_specific(types.Region.OVERSEAS)
    async def get_mimo_lottery_info(
        self,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> models.MimoLotteryInfo:
        """Get Traveling Mimo lottery info."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        data = await self._request_mimo(
            "lottery-info",
            params=dict(game_id=game_id, lang=lang or self.lang, version_id=version_id),
        )
        return models.MimoLotteryInfo(**data)

    @base.region_specific(types.Region.OVERSEAS)
    async def draw_mimo_lottery(
        self,
        *,
        game_id: typing.Optional[int] = None,
        version_id: typing.Optional[int] = None,
        game: typing.Optional[typing.Union[typing.Literal["hoyolab"], types.Game]] = None,
        lang: typing.Optional[str] = None,
    ) -> models.MimoLotteryResult:
        """Draw a Traveling Mimo lottery."""
        game_id, version_id = await self._parse_mimo_args(game_id, version_id, game)
        data = await self._request_mimo(
            "lottery",
            data=dict(game_id=game_id, lang=lang or self.lang, version_id=version_id),
            method="POST",
        )
        return models.MimoLotteryResult(**data)

    async def reply_to_post(self, content: str, *, post_id: int) -> int:
        """Reply to a community post."""
        data = await self.request_bbs(
            "community/post/wapi/releaseReply",
            data=dict(
                post_id=str(post_id),
                content=f"<p>{content}</p>",
                image_list=[],
                reply_bubble_id="",
                structured_content=json.dumps([{"insert": f"{content}\n"}]),
            ),
            method="POST",
            headers={"x-rpc-device_id": str(uuid.uuid4())},
        )
        return int(data["reply_id"])

    async def delete_reply(self, *, reply_id: int, post_id: int) -> None:
        """Delete a reply."""
        await self.request_bbs(
            "community/post/wapi/deleteReply",
            data=dict(reply_id=reply_id, post_id=post_id),
            method="POST",
        )

    async def _request_join(self, topic_id: int, *, is_cancel: bool) -> None:
        await self.request_bbs(
            "community/topic/wapi/join",
            data=dict(topic_id=topic_id, is_cancel=is_cancel),
            method="POST",
        )

    async def join_topic(self, topic_id: int) -> None:
        """Join a topic."""
        await self._request_join(topic_id, is_cancel=False)

    async def leave_topic(self, topic_id: int) -> None:
        """Leave a topic."""
        await self._request_join(topic_id, is_cancel=True)

    @base.region_specific(types.Region.OVERSEAS)
    async def get_web_events(
        self,
        game: typing.Optional[types.Game] = None,
        *,
        size: int = 15,
        offset: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> list[models.WebEvent]:
        """Get a list of web events."""
        game = game or self.default_game
        if game is None:
            raise ValueError("No default game set.")

        data = await self.request_bbs(
            "community/community_contribution/wapi/event/list",
            params=dict(gids=constants.WEB_EVENT_GAME_IDS[game], size=size, offset=offset or ""),
            lang=lang,
        )
        return [models.WebEvent(**i) for i in data["list"]]

    @base.region_specific(types.Region.OVERSEAS)
    async def get_accompany_characters(
        self, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[models.AccompanyCharacterGame]:
        """Get a list of accompany characters, this endpoint doesn't require cookies."""
        data = await self.request_bbs(
            "community/painter/api/getChannelRoleList",
            cache=client_cache.cache_key("accp_chars"),
            method="POST",
            lang=lang,
        )
        return [models.AccompanyCharacterGame(**i) for i in data["game_roles_list"]]

    @base.region_specific(types.Region.OVERSEAS)
    async def accompany_character(self, *, role_id: int, topic_id: int) -> models.AccompanyResult:
        """Accompany a character, role_id and topic_id can be found by calling get_accompany_characters."""
        data = await self.request_bbs(
            "community/apihub/api/user/accompany/role", params=dict(role_id=role_id, topic_id=topic_id)
        )
        return models.AccompanyResult(**data)



"""Errors received from the API."""

import typing

from genshin.constants import GEETEST_RETCODES

__all__ = [
    "ERRORS",
    "AccountMuted",
    "AccountNotFound",
    "AlreadyClaimed",
    "AuthkeyException",
    "AuthkeyTimeout",
    "CookieException",
    "DailyGeetestTriggered",
    "DataNotPublic",
    "GeetestError",
    "GenshinException",
    "InvalidAuthkey",
    "InvalidCookies",
    "RedemptionClaimed",
    "RedemptionCooldown",
    "RedemptionException",
    "RedemptionInvalid",
    "TooManyRequests",
    "check_for_geetest",
    "raise_for_retcode",
]


class GenshinException(Exception):
    """A base genshin exception."""

    retcode: int = 0
    original: str = ""
    msg: str = ""

    def __init__(
        self, response: typing.Optional[typing.Mapping[str, typing.Any]] = None, msg: typing.Optional[str] = None
    ) -> None:
        response = response or {}

        self.retcode = response.get("retcode", self.retcode)
        self.original = response.get("message", "")
        self.msg = msg or self.msg or self.original

        if self.retcode and f"[{self.retcode}]" not in self.msg:
            msg = f"[{self.retcode}] {self.msg}"
        else:
            msg = self.msg

        super().__init__(msg)

    def __repr__(self) -> str:
        response = {"retcode": self.retcode, "message": self.original}
        args = [repr(response)]
        if self.msg != self.original:
            args.append(repr(self.msg))

        return f"{type(self).__name__}({', '.join(args)})"

    @property
    def response(self) -> typing.Mapping[str, typing.Any]:
        return {"retcode": self.retcode, "message": self.original, "data": None}


class InternalDatabaseError(GenshinException):
    """Internal database error."""

    retcode = -1


class AccountNotFound(GenshinException):
    """Tried to get data with an invalid uid."""

    msg = "Could not find user; uid may be invalid."


class DataNotPublic(GenshinException):
    """User hasn't set their data to public."""

    msg = "User's data is not public."


class CookieException(GenshinException):
    """Base error for cookies."""


class InvalidCookies(CookieException):
    """Cookies weren't valid."""

    retcode = -100
    msg = "Cookies are not valid."


class TooManyRequests(CookieException):
    """Made too many requests and got ratelimited."""

    retcode = 10101
    msg = "Cannot get data for more than 30 accounts per cookie per day."


class VisitsTooFrequently(GenshinException):
    """Visited a page too frequently.

    Must be handled with exponential backoff.
    """

    retcode = -110
    msg = "Visits too frequently."


class AlreadyClaimed(GenshinException):
    """Already claimed the daily reward today."""

    retcode = -5003
    msg = "Already claimed the daily reward today."


class DailyGeetestTriggered(GenshinException):
    """Geetest triggered during daily reward claim."""

    msg = "Geetest triggered during daily reward claim."

    gt: str
    challenge: str

    def __init__(self, response: typing.Mapping[str, typing.Any], *, gt: str, challenge: str) -> None:
        self.gt = gt
        self.challenge = challenge

        super().__init__(response)


class AuthkeyException(GenshinException):
    """Base error for authkeys."""


class InvalidAuthkey(AuthkeyException):
    """Authkey is not valid."""

    retcode = -100
    msg = "Authkey is not valid."


class AuthkeyTimeout(AuthkeyException):
    """Authkey has timed out."""

    retcode = -101
    msg = "Authkey has timed out."


class RedemptionException(GenshinException):
    """Exception caused by redeeming a code."""


class RedemptionInvalid(RedemptionException):
    """Invalid redemption code."""

    msg = "Invalid redemption code."


class RedemptionCooldown(RedemptionException):
    """Redemption is on cooldown."""

    msg = "Redemption is on cooldown."


class RedemptionClaimed(RedemptionException):
    """Redemption code has been claimed already."""

    msg = "Redemption code has been claimed already."


class RedeemGameLevelTooLow(RedemptionException):
    """Redemption code cannot be claimed because the game level is too low."""


class AccountLoginFail(GenshinException):
    """Account if not exists in hoyoverse (Or password incorrect)."""

    msg = "Account login failed."


class AccountHasLocked(GenshinException):
    """Account has logged incorrect over than 3 - 5 time(s). It's will be locked and wait 20 minute."""

    msg = "Account has been locked because exceeded password limit. Please wait 20 minute and try again"


class WrongOTP(GenshinException):
    """Wrong OTP code."""

    msg = "The provided OTP code is wrong."


class GeetestError(GenshinException):
    """Geetest triggered during the battle chronicle API request."""

    def __init__(self, response: dict[str, typing.Any]) -> None:
        super().__init__(response)

    msg = "Geetest triggered during the battle chronicle API request."


class OTPRateLimited(GenshinException):
    """Too many OTP messages sent for the number.

    The limit is 40 messages/day/number.
    """

    retcode = -119
    msg = "Too many OTP messages sent for the number."


class IncorrectGameAccount(GenshinException):
    """Game account is incorrect."""

    retcode = -216
    msg = "Game account is incorrect."


class IncorrectGamePassword(GenshinException):
    """Game password is incorrect."""

    retcode = -202
    msg = "Game password is incorrect."


class AccountDoesNotExist(GenshinException):
    """Account does not exist."""

    retcode = -3203
    msg = "Account does not exist."


class VerificationCodeRateLimited(GenshinException):
    """Too many verification code requests for the account."""

    retcode = -3206
    msg = "Too many verification code requests for the account."


class AccountMuted(GenshinException):
    """Account is muted."""

    retcode = 2010
    msg = "Account is muted."


class ActionInCooldown(GenshinException):
    """Action is in cooldown."""


class NoNeedGeetest(GenshinException):
    """No need to do geetest."""


_TGE = type[GenshinException]
_errors: dict[int, typing.Union[_TGE, str, tuple[_TGE, typing.Optional[str]]]] = {
    # misc hoyolab
    -100: InvalidCookies,
    -108: "Invalid language.",
    -110: VisitsTooFrequently,
    1028: VisitsTooFrequently,
    2010: AccountMuted,
    # game record
    10001: InvalidCookies,
    -10001: "Malformed request.",
    -10002: "No game account associated with cookies.",
    # database game record
    10101: TooManyRequests,
    10102: DataNotPublic,
    10103: (InvalidCookies, "Cookies are valid but do not have a hoyolab account bound to them."),
    10104: "Cannot view real-time notes of other users.",
    # calculator
    -500001: "Invalid fields in calculation.",
    -500004: VisitsTooFrequently,
    -502001: "User does not have this character.",
    -502002: "Calculator sync is not enabled.",
    # mixin
    -1: InternalDatabaseError,
    10307: InternalDatabaseError,
    1009: AccountNotFound,
    # redemption
    -1065: RedemptionInvalid,
    -1071: InvalidCookies,
    -1073: (AccountNotFound, "Account has no game account bound to it."),
    -2001: (RedemptionInvalid, "Redemption code has expired."),
    -2003: (RedemptionInvalid, "Redemption code is incorrectly formatted."),
    -2004: RedemptionInvalid,
    -2006: (RedemptionInvalid, "Redemption code has reached max usage limit."),
    -2014: (RedemptionInvalid, "Redemption code not activated"),
    -2016: RedemptionCooldown,
    -2017: RedemptionClaimed,
    -2018: RedemptionClaimed,
    -2021: RedeemGameLevelTooLow,
    -2011: RedeemGameLevelTooLow,
    # rewards
    -5003: AlreadyClaimed,
    # chinese
    1008: AccountNotFound,
    -1104: "This action must be done in the app.",
    # account
    -3004: AccountLoginFail,
    -3208: AccountLoginFail,
    -3202: AccountHasLocked,
    -3203: AccountDoesNotExist,
    -3205: WrongOTP,
    -3206: VerificationCodeRateLimited,
    -3102: WrongOTP,
    # Miyoushe
    -119: OTPRateLimited,
    -3006: "Request too frequent.",  # OTP endpoint
    # Game login
    -216: IncorrectGameAccount,
    -202: IncorrectGamePassword,
    # lineup
    -1004: ActionInCooldown,
    -3101: ActionInCooldown,
    # geetest
    30001: NoNeedGeetest,
}

ERRORS: dict[int, tuple[_TGE, typing.Optional[str]]] = {
    retcode: (GenshinException, exc) if isinstance(exc, str) else exc if isinstance(exc, tuple) else (exc, None)
    for retcode, exc in _errors.items()
}


def raise_for_retcode(data: dict[str, typing.Any]) -> typing.NoReturn:
    """Raise an equivalent error to a response.

    game record:
        10001 = invalid cookie
        101xx = generic errors
    authkey:
        -100 = invalid authkey
        -101 = authkey timed out
    code redemption:
        20xx = invalid code or state
        -107x = invalid cookies
    daily reward:
        -500x = already claimed the daily reward
    """
    r, m = data.get("retcode", 0), data.get("message", "")

    if m.startswith("authkey"):
        if r == -100:
            raise InvalidAuthkey(data)
        elif r == -101:
            raise AuthkeyTimeout(data)
        else:
            raise AuthkeyException(data)

    if r in ERRORS:
        exctype, msg = ERRORS[r]
        raise exctype(data, m or msg)

    if "redemption" in m:
        raise RedemptionException(data)

    raise GenshinException(data)


def check_for_geetest(data: dict[str, typing.Any]) -> None:
    """Check if geetest was triggered during the request and raise an error if so."""
    retcode = data.get("retcode")
    if retcode is not None and retcode in GEETEST_RETCODES:
        raise GeetestError(data)

    if not data.get("data"):  # if is an error
        return

    gt_result: typing.Any = data["data"].get("gt_result", data["data"])

    if not gt_result:
        return

    if (
        gt_result.get("risk_code") != 0
        and gt_result.get("gt")
        and gt_result.get("challenge")
        and gt_result.get("success") != 0
    ):
        raise DailyGeetestTriggered(data, gt=gt_result.get("gt", ""), challenge=gt_result.get("challenge", ""))



"""Modern API wrapper for Genshin Impact & Honkai Impact 3rd built on asyncio and pydantic.

Documentation: https://gh.seria.moe/genshin.py

Source Code: https://github.com/seriaati/genshin.py
"""

from . import models, utility
from .client import *
from .constants import *
from .errors import *
from .types import *

__version__ = "1.0.0"



"""Diary component."""

import datetime
import functools
import typing

from genshin import paginators, types, utility
from genshin.client import cache, routes
from genshin.client.components import base
from genshin.client.manager import managers
from genshin.constants import CN_TIMEZONE
from genshin.models.genshin import diary as models
from genshin.utility import deprecation

__all__ = ["DiaryClient"]


class DiaryCallback(typing.Protocol):
    """Callback which requires a diary page."""

    async def __call__(self, page: int, /) -> models.DiaryPage:
        """Return a diary page."""
        ...


class DiaryPaginator(paginators.PagedPaginator[models.DiaryAction]):
    """Paginator for diary."""

    _data: typing.Optional[models.DiaryPage]
    """Metadata of the paginator"""

    def __init__(self, getter: DiaryCallback, *, limit: typing.Optional[int] = None) -> None:
        self._get_page = getter
        self._data = None

        super().__init__(self._getter, limit=limit, page_size=100)

    async def _getter(self, page: int) -> typing.Sequence[models.DiaryAction]:
        self._data = await self._get_page(page)
        return self._data.actions

    @property
    def data(self) -> models.BaseDiary:
        """Get data bound to the diary.

        This requires at least one page to have been fetched.
        """
        if self._data is None:
            raise RuntimeError("At least one item must be fetched before data can be gotten.")

        return self._data


class StarRailDiaryCallback(typing.Protocol):
    """Callback which requires a diary page."""

    async def __call__(self, page: int, /) -> models.StarRailDiaryPage:
        """Return a diary page."""
        ...


class StarRailDiaryPaginator(paginators.PagedPaginator[models.StarRailDiaryAction]):
    """Paginator for diary."""

    _data: typing.Optional[models.StarRailDiaryPage]
    """Metadata of the paginator"""

    def __init__(self, getter: StarRailDiaryCallback, *, limit: typing.Optional[int] = None) -> None:
        self._get_page = getter
        self._data = None

        super().__init__(self._getter, limit=limit, page_size=100)

    async def _getter(self, page: int) -> typing.Sequence[models.StarRailDiaryAction]:
        self._data = await self._get_page(page)
        return self._data.actions

    @property
    def data(self) -> models.BaseDiary:
        """Get data bound to the diary.

        This requires at least one page to have been fetched.
        """
        if self._data is None:
            raise RuntimeError("At least one item must be fetched before data can be gotten.")

        return self._data


class DiaryClient(base.BaseClient):
    """Diary component."""

    @managers.no_multi
    async def request_ledger(
        self,
        uid: typing.Optional[int] = None,
        *,
        game: typing.Optional[types.Game] = None,
        detail: bool = False,
        month: typing.Union[int, str, None] = None,
        lang: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the ys ledger endpoint."""
        # TODO: Do not separate urls?
        params = dict(params or {})

        if game is None:
            if self.default_game is None:
                raise RuntimeError("No default game set.")

            game = self.default_game

        base_url = routes.DETAIL_LEDGER_URL if detail else routes.INFO_LEDGER_URL
        url = base_url.get_url(self.region, game)

        uid = uid or await self._get_uid(game)

        if self.region == types.Region.OVERSEAS or game == types.Game.STARRAIL:
            params["uid"] = uid
            params["region"] = utility.recognize_server(uid, game)
        elif self.region == types.Region.CHINESE:
            params["bind_uid"] = uid
            params["bind_region"] = utility.recognize_server(uid, game)
        else:
            raise TypeError(f"{self.region!r} is not a valid region.")
        params["month"] = month or (
            datetime.datetime.now().strftime("%Y%m") if game == types.Game.STARRAIL else datetime.datetime.now().month
        )
        params["lang"] = lang or self.lang

        return await self.request(url, params=params, **kwargs)

    @deprecation.deprecated("get_genshin_diary")
    async def get_diary(
        self,
        uid: typing.Optional[int] = None,
        *,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.Diary:
        """Get a traveler's diary with earning details for the month."""
        return await self.get_genshin_diary(uid, month=month, lang=lang)

    async def get_genshin_diary(
        self,
        uid: typing.Optional[int] = None,
        *,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.Diary:
        """Get a traveler's diary with earning details for the month."""
        game = types.Game.GENSHIN
        uid = uid or await self._get_uid(game)
        cache_key = cache.cache_key(
            "diary", uid=uid, game=game, month=month or datetime.datetime.now(CN_TIMEZONE).month, lang=lang or self.lang
        )
        data = await self.request_ledger(uid, game=game, month=month, lang=lang, cache=cache_key)
        return models.Diary(**data)

    async def get_starrail_diary(
        self,
        uid: typing.Optional[int] = None,
        *,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.StarRailDiary:
        """Get a blazer's diary with earning details for the month."""
        game = types.Game.STARRAIL
        uid = uid or await self._get_uid(game)
        cache_key = cache.cache_key(
            "diary", uid=uid, game=game, month=month or datetime.datetime.now(CN_TIMEZONE).month, lang=lang or self.lang
        )
        data = await self.request_ledger(uid, game=game, month=month, lang=lang, cache=cache_key)
        return models.StarRailDiary(**data)

    async def _get_genshin_diary_page(
        self,
        page: int,
        *,
        uid: typing.Optional[int] = None,
        type: int = models.DiaryType.PRIMOGEMS,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.DiaryPage:
        data = await self.request_ledger(
            uid,
            game=types.Game.GENSHIN,
            detail=True,
            month=month,
            lang=lang,
            params=dict(type=type, current_page=page, page_size=100),
        )
        return models.DiaryPage(**data)

    @deprecation.deprecated("genshin_diary_log")
    def diary_log(
        self,
        uid: typing.Optional[int] = None,
        *,
        limit: typing.Optional[int] = None,
        type: int = models.DiaryType.PRIMOGEMS,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> DiaryPaginator:
        """Create a new daily reward paginator."""
        return self.genshin_diary_log(
            uid=uid,
            limit=limit,
            type=type,
            month=month,
            lang=lang,
        )

    def genshin_diary_log(
        self,
        uid: typing.Optional[int] = None,
        *,
        limit: typing.Optional[int] = None,
        type: int = models.DiaryType.PRIMOGEMS,
        month: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> DiaryPaginator:
        """Create a new daily reward paginator."""
        return DiaryPaginator(
            functools.partial(
                self._get_genshin_diary_page,
                uid=uid,
                type=type,
                month=month,
                lang=lang,
            ),
            limit=limit,
        )

    async def _get_starrail_diary_page(
        self,
        page: int,
        *,
        uid: typing.Optional[int] = None,
        type: int = models.StarRailDiaryType.STELLARJADE,
        month: typing.Optional[str] = None,
        lang: typing.Optional[str] = None,
    ) -> models.StarRailDiaryPage:
        data = await self.request_ledger(
            uid,
            game=types.Game.STARRAIL,
            detail=True,
            month=month,
            lang=lang,
            params=dict(type=type, current_page=page, page_size=100),
        )
        return models.StarRailDiaryPage(**data)

    def starrail_diary_log(
        self,
        uid: typing.Optional[int] = None,
        *,
        limit: typing.Optional[int] = None,
        type: int = models.StarRailDiaryType.STELLARJADE,
        month: typing.Optional[str] = None,
        lang: typing.Optional[str] = None,
    ) -> StarRailDiaryPaginator:
        """Create a new daily reward paginator."""
        return StarRailDiaryPaginator(
            functools.partial(
                self._get_starrail_diary_page,
                uid=uid,
                type=type,
                month=month,
                lang=lang,
            ),
            limit=limit,
        )



import typing

import aiohttp.typedefs

from genshin import types
from genshin.client import routes
from genshin.client.components import base
from genshin.constants import GAME_LANGS
from genshin.models.starrail import rpgsimulator as models
from genshin.utility import ds
from genshin.utility.uid import recognize_server

__all__ = ("HSRLineupClient",)

LineupGameMode = typing.Union[models.StarRailGameModeType, typing.Literal["Chasm", "Story", "Boss"]]


class HSRLineupClient(base.BaseClient):
    """HSR lineup simulator client."""

    async def _request(
        self,
        endpoint: str,
        *,
        lang: typing.Optional[str] = None,
        region: typing.Optional[types.Region] = None,
        method: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request any hoyolab endpoint."""
        if lang is not None and lang not in GAME_LANGS[types.Game.STARRAIL]:
            raise ValueError(f"Invalid language {lang!r} for game {types.Game.STARRAIL}.")

        lang = lang or self.lang
        region = region or self.region

        url = routes.HSR_LINEUP_SIMULATOR_URL.get_url() / endpoint

        headers = base.parse_loose_headers(headers)
        headers.update(ds.get_ds_headers(data=data, params=params, region=region, lang=lang))

        return await self.request(url, method=method, params=params, data=data, headers=headers, **kwargs)

    async def get_starrail_lineup_game_modes(
        self, *, lang: typing.Optional[str] = None
    ) -> list[models.StarRailGameMode]:
        """Get the available game modes for the HSR lineup simulator."""
        data = await self._request("tag", lang=lang)
        return [models.StarRailGameMode(**item) for item in data["tree"]]

    def get_starrail_lineup_floor(
        self,
        game_modes: typing.Sequence[models.StarRailGameMode],
        *,
        type: LineupGameMode,
        floor: int,
    ) -> typing.Optional[models.StarRailGameModeFloor]:
        """Get a specific floor from the game modes."""
        for mode in game_modes:
            if mode.type != type:
                continue
            for f in mode.floors:
                if f.floor == floor:
                    return f

        return None

    @typing.overload
    async def get_starrail_lineups(
        self,
        *,
        tag_id: int,
        group_id: int,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.MOC], typing.Literal["Chasm"]],
        next_page_token: typing.Optional[str] = ...,
        order: typing.Literal["Hot", "Match", "CreatedTime"] = ...,
        lang: typing.Optional[str] = ...,
    ) -> models.StarRailLineupResponse: ...
    @typing.overload
    async def get_starrail_lineups(
        self,
        *,
        tag_id: int,
        group_id: int,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.PURE_FICTION], typing.Literal["Story"]],
        next_page_token: typing.Optional[str] = ...,
        order: typing.Literal["Hot", "Match", "CreatedTime"] = ...,
        lang: typing.Optional[str] = ...,
    ) -> models.PureFictionLineupResponse: ...
    @typing.overload
    async def get_starrail_lineups(
        self,
        *,
        tag_id: int,
        group_id: int,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.APC_SHADOW], typing.Literal["Boss"]],
        next_page_token: typing.Optional[str] = ...,
        order: typing.Literal["Hot", "Match", "CreatedTime"] = ...,
        lang: typing.Optional[str] = ...,
    ) -> models.APCShadowLineupResponse: ...
    async def get_starrail_lineups(
        self,
        *,
        tag_id: int,
        group_id: int,
        type: LineupGameMode,
        next_page_token: typing.Optional[str] = None,
        order: typing.Literal["Hot", "Match", "CreatedTime"] = "Match",
        lang: typing.Optional[str] = None,
    ) -> typing.Union[models.StarRailLineupResponse, models.PureFictionLineupResponse, models.APCShadowLineupResponse]:
        """Get the available lineups for the HSR lineup simulator."""
        params: dict[str, typing.Any] = {
            "tag_id": tag_id,
            "group_id": group_id,
            "lineup_type": str(type),
            "game": "hkrpg",
        }

        if order not in {"Hot", "Match", "CreatedTime"}:
            msg = f"Invalid order {order!r}. Must be one of 'Hot', 'Match', or 'CreatedTime'."
            raise ValueError(msg)
        params["order"] = order

        if self.uid is not None:
            params["uid"] = self.uid
            params["region"] = recognize_server(self.uid, types.Game.STARRAIL)

        if next_page_token:
            params["next_page_token"] = next_page_token

        data = await self._request("lineup/index", lang=lang, params=params)

        if type == "Story":
            return models.PureFictionLineupResponse(**data)
        elif type == "Boss":
            return models.APCShadowLineupResponse(**data)
        return models.StarRailLineupResponse(**data)

    @typing.overload
    async def get_starrail_lineup_schedules(
        self,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.MOC], typing.Literal["Chasm"]],
        *,
        lang: typing.Optional[str] = ...,
    ) -> list[models.MOCSchedule]: ...
    @typing.overload
    async def get_starrail_lineup_schedules(
        self,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.PURE_FICTION], typing.Literal["Story"]],
        *,
        lang: typing.Optional[str] = ...,
    ) -> list[models.PureFictionSchedule]: ...
    @typing.overload
    async def get_starrail_lineup_schedules(
        self,
        type: typing.Union[typing.Literal[models.StarRailGameModeType.APC_SHADOW], typing.Literal["Boss"]],
        *,
        lang: typing.Optional[str] = ...,
    ) -> list[models.APCShadowSchedule]: ...
    async def get_starrail_lineup_schedules(
        self, type: LineupGameMode, *, lang: typing.Optional[str] = None
    ) -> typing.Union[list[models.MOCSchedule], list[models.PureFictionSchedule], list[models.APCShadowSchedule]]:
        """Get the schedule for the HSR lineup simulator."""
        endpoints = {
            "Chasm": "schedule/list",
            "Story": "story_schedule/list",
            "Boss": "boss_schedule/list",
        }
        endpoint = endpoints.get(type)
        if endpoint is None:
            msg = f"Invalid type {type!r} for HSR lineup schedule."
            raise ValueError(msg)

        data = await self._request(endpoint, lang=lang)
        if type == "Boss":
            return [models.APCShadowSchedule(**item) for item in data["schedule"]]
        if type == "Chasm":
            return [models.MOCSchedule(**item) for item in data["schedule"]]
        return [models.PureFictionSchedule(**item) for item in data["schedule"]]



"""Calculator client."""

from __future__ import annotations

import asyncio
import typing
import warnings

import aiohttp.typedefs

import genshin.models.genshin as genshin_models
from genshin import errors, types, utility
from genshin.client import cache as client_cache
from genshin.client import routes
from genshin.client.components import base
from genshin.models.genshin import calculator as models
from genshin.utility import deprecation

from .calculator import BatchCalculator, Calculator, FurnishingCalculator

__all__ = ["CalculatorClient"]


class CalculatorClient(base.BaseClient):
    """Calculator component."""

    async def request_calculator(
        self,
        endpoint: str,
        *,
        method: str = "POST",
        lang: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        api_version: int = 1,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the calculator endpoint."""
        params = dict(params or {})
        headers = base.parse_loose_headers(headers)

        base_url = routes.CALCULATOR_URL.get_url(self.region)
        url = base_url / f"v{api_version}" / endpoint

        if method == "GET":
            params["lang"] = lang or self.lang
            data = None
        else:
            data = dict(data or {})
            data["lang"] = lang or self.lang

        headers["referer"] = str(routes.CALCULATOR_REFERER_URL.get_url(self.region))
        update_task = asyncio.create_task(utility.update_characters_any(lang or self.lang, lenient=True))

        data = await self.request(url, method=method, params=params, data=data, headers=headers, **kwargs)

        try:
            await update_task
        except Exception as e:
            warnings.warn(f"Failed to update characters: {e!r}")

        return data

    async def _execute_calculator(
        self,
        data: typing.Mapping[str, typing.Any],
        *,
        lang: typing.Optional[str] = None,
    ) -> models.CalculatorResult:
        """Calculate the results of a builder."""
        api_data = await self.request_calculator("compute", lang=lang, data=data, api_version=2)
        return models.CalculatorResult(**api_data)

    async def _execute_batch_calculator(
        self,
        data: typing.Sequence[typing.Mapping[str, typing.Any]],
        *,
        lang: typing.Optional[str] = None,
    ) -> models.CalculatorBatchResult:
        """Calculate the results of a batch builder."""
        api_data = await self.request_calculator("batch_compute", lang=lang, data={"items": data}, api_version=3)
        return models.CalculatorBatchResult(**api_data)

    async def _execute_furnishings_calculator(
        self,
        data: typing.Mapping[str, typing.Any],
        *,
        lang: typing.Optional[str] = None,
    ) -> models.CalculatorFurnishingResults:
        """Calculate the results of a builder."""
        data = await self.request_calculator("furniture/compute", lang=lang, data=data)
        return models.CalculatorFurnishingResults(**data)

    def calculator(self, *, lang: typing.Optional[str] = None) -> Calculator:
        """Create a calculator builder object."""
        return Calculator(self, lang=lang)

    def batch_calculator(self, *, lang: typing.Optional[str] = None) -> BatchCalculator:
        """Create a batch calculator builder object."""
        return BatchCalculator(self, lang=lang)

    def furnishings_calculator(self, *, lang: typing.Optional[str] = None) -> FurnishingCalculator:
        """Create a calculator builder object."""
        return FurnishingCalculator(self, lang=lang)

    async def _enable_calculator_sync(self, enabled: bool = True) -> None:
        """Enable data syncing in calculator."""
        await self.request_calculator("avatar/auth", method="POST", data=dict(avatar_auth=int(enabled)))

    async def _get_calculator_items(
        self,
        slug: str,
        filters: typing.Mapping[str, typing.Any],
        query: typing.Optional[str] = None,
        *,
        uid: typing.Optional[int] = None,
        is_all: bool = False,
        sync: bool = False,
        lang: typing.Optional[str] = None,
        autoauth: bool = True,
    ) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
        """Get all items of a specific slug from a calculator."""
        endpoint = f"sync/{slug}/list" if sync else f"{slug}/list"

        if query:
            if any(filters.values()):
                raise TypeError("Cannot specify a query and filter at the same time")

            filters = dict(keywords=query, **filters)

        payload: dict[str, typing.Any] = dict(page=1, size=200, **filters)

        if sync:
            uid = uid or await self._get_uid(types.Game.GENSHIN)
            payload["uid"] = uid
            payload["region"] = utility.recognize_genshin_server(uid)

        cache: typing.Optional[client_cache.CacheKey] = None
        if not any(filters.values()) and not sync:
            cache = client_cache.cache_key("calculator", slug=slug, lang=lang or self.lang)

        try:
            data = await self.request_calculator(endpoint, lang=lang, data=payload, cache=cache)
        except errors.GenshinException as e:
            if e.retcode != -502002:  # Sync not enabled
                raise
            if not autoauth:
                raise errors.GenshinException(e.response, "Calculator sync is not enabled") from e

            await self._enable_calculator_sync()
            data = await self.request_calculator(endpoint, lang=lang, data=payload, cache=cache)

        return data["list"]

    async def get_calculator_characters(
        self,
        *,
        query: typing.Optional[str] = None,
        elements: typing.Optional[typing.Sequence[int]] = None,
        weapon_types: typing.Optional[typing.Sequence[int]] = None,
        include_traveler: bool = False,
        sync: bool = False,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorCharacter]:
        """Get all characters provided by the Enhancement Progression Calculator."""
        data = await self._get_calculator_items(
            "avatar",
            lang=lang,
            is_all=include_traveler,
            sync=sync,
            uid=uid,
            query=query,
            filters=dict(
                element_attr_ids=elements or [],
                weapon_cat_ids=weapon_types or [],
            ),
        )
        return [models.CalculatorCharacter(**i) for i in data]

    async def get_calculator_weapons(
        self,
        *,
        query: typing.Optional[str] = None,
        types: typing.Optional[typing.Sequence[int]] = None,
        rarities: typing.Optional[typing.Sequence[int]] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorWeapon]:
        """Get all weapons provided by the Enhancement Progression Calculator."""
        data = await self._get_calculator_items(
            "weapon",
            lang=lang,
            query=query,
            filters=dict(
                weapon_cat_ids=types or [],
                weapon_levels=rarities or [],
            ),
        )
        return [models.CalculatorWeapon(**i) for i in data]

    async def get_calculator_artifacts(
        self,
        *,
        query: typing.Optional[str] = None,
        pos: int = 1,
        rarities: typing.Optional[typing.Sequence[int]] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorArtifact]:
        """Get all artifacts provided by the Enhancement Progression Calculator."""
        data = await self._get_calculator_items(
            "reliquary",
            lang=lang,
            query=query,
            filters=dict(
                reliquary_cat_id=pos,
                reliquary_levels=rarities or [],
            ),
        )
        return [models.CalculatorArtifact(**i) for i in data]

    async def get_calculator_furnishings(
        self,
        *,
        types: typing.Optional[int] = None,
        rarities: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorFurnishing]:
        """Get all furnishings provided by the Enhancement Progression Calculator."""
        data = await self._get_calculator_items(
            "furniture",
            lang=lang,
            filters=dict(
                cat_id=types or 0,
                weapon_levels=rarities or 0,
            ),
        )
        return [models.CalculatorFurnishing(**i) for i in data]

    async def get_character_details(
        self,
        character: types.IDOr[genshin_models.BaseCharacter],
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.CalculatorCharacterDetails:
        """Get the weapon, artifacts and talents of a character.

        Not related to the Battle Chronicle.
        This data is always private.
        """
        uid = uid or await self._get_uid(types.Game.GENSHIN)

        data = await self.request_calculator(
            "sync/avatar/detail",
            method="GET",
            lang=lang,
            params=dict(
                avatar_id=int(character),
                uid=uid,
                region=utility.recognize_genshin_server(uid),
            ),
        )
        return models.CalculatorCharacterDetails(**data)

    async def get_complete_artifact_set(
        self,
        artifact: types.IDOr[typing.Union[genshin_models.Artifact, genshin_models.CalculatorArtifact]],
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorArtifact]:
        """Get all other artifacts that share a set with any given artifact.

        Doesn't return the artifact passed into this function.
        """
        data = await self.request_calculator(
            "reliquary/set",
            method="GET",
            lang=lang,
            params=dict(reliquary_id=int(artifact)),
            cache=client_cache.cache_key("calculator", slug="set", artifact=int(artifact), lang=lang or self.lang),
        )
        return [models.CalculatorArtifact(**i) for i in data["reliquary_list"]]

    async def _get_all_artifact_ids(self, artifact_id: int) -> typing.Sequence[int]:
        """Get all artifact ids in the same set as a given artifact id."""
        others = await self.get_complete_artifact_set(artifact_id)
        return [artifact_id] + [other.id for other in others]

    async def get_teapot_replica_blueprint(
        self,
        share_code: int,
        *,
        region: typing.Optional[str] = None,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.CalculatorFurnishing]:
        """Get furnishings used by a teapot replica blueprint."""
        if not region:
            uid = uid or await self._get_uid(types.Game.GENSHIN)
            region = utility.recognize_genshin_server(uid)

        data = await self.request_calculator(
            "furniture/blueprint",
            method="GET",
            lang=lang,
            params=dict(share_code=share_code, region=region),
            cache=client_cache.cache_key("calculator", slug="blueprint", share_code=share_code, lang=lang or self.lang),
        )
        return [models.CalculatorFurnishing(**i) for i in data["list"]]

    @deprecation.deprecated("await genshin.utility.update_characters_any()")
    async def update_character_names(self, *, lang: typing.Optional[str] = None) -> None:
        """Update stored db characters with the names from the calculator."""
        characters = await self.get_calculator_characters(lang=lang, include_traveler=True)

        for char in characters:
            icon = genshin_models.character._parse_icon(char.icon)
            dbchar = genshin_models.DBChar(
                char.id, icon, char.name, "" if "Player" in icon else char.element, char.rarity
            )

            genshin_models.CHARACTER_NAMES[lang or self.lang][dbchar.id] = dbchar



"""Starrail chronicle character."""

import typing
from collections.abc import Mapping, Sequence

import pydantic

from genshin.models.model import APIModel

from .. import character

__all__ = [
    "ModifyRelicProperty",
    "RecommendProperty",
    "StarRailDetailCharacterResponse",
    "StarRailSimpleCharacterResponse",
]


class RecommendProperty(APIModel):
    """Character recommended and preferred properties."""

    recommend_relic_properties: typing.Sequence[int]
    custom_relic_properties: typing.Sequence[int]
    is_custom_property_valid: bool


class ModifyRelicProperty(APIModel):
    """Modify relic property."""

    property_type: int
    modify_property_type: int


class StarRailSimpleCharacterResponse(APIModel):
    """HSR characters endpoint response model for when viewed by other players."""

    avatar_list: Sequence[character.StarRailSimpleCharacter]
    equip_wiki: Mapping[str, str]
    relic_wiki: Mapping[str, str]
    property_info: Mapping[str, character.PropertyInfo]
    recommend_property: Mapping[str, RecommendProperty]
    relic_properties: Sequence[ModifyRelicProperty]

    @pydantic.model_validator(mode="before")
    def __fill_additional_fields(cls, values: Mapping[str, typing.Any]) -> Mapping[str, typing.Any]:
        """Fill additional fields for convenience."""
        characters = values.get("avatar_list", [])
        props_info = values.get("property_info", {})
        rec_props = values.get("recommend_property", {})
        equip_wiki = values.get("equip_wiki", {})
        relic_wiki = values.get("relic_wiki", {})

        for char in characters:
            char_id = str(char["id"])
            char_rec_props = rec_props[char_id]["recommend_relic_properties"]
            char_custom_props = rec_props[char_id]["custom_relic_properties"]

            servant = char.get("servant_detail", {})
            if servant:
                for prop in servant["servant_properties"]:
                    prop_type = prop["property_type"]
                    prop["info"] = props_info[str(prop_type)]

            for relic in char["relics"] + char["ornaments"]:
                if relic["main_property"] is None:
                    # Relic has no main stat and substats when the profile is viewed by other players
                    continue

                prop_type = relic["main_property"]["property_type"]
                relic["main_property"]["info"] = props_info[str(prop_type)]
                relic["main_property"]["recommended"] = prop_type in char_rec_props
                relic["main_property"]["preferred"] = prop_type in char_custom_props

                for prop in relic["properties"]:
                    prop_type = prop["property_type"]
                    prop["recommended"] = prop_type in char_rec_props
                    prop["preferred"] = prop_type in char_custom_props
                    prop["info"] = props_info[str(prop_type)]

                relic["wiki"] = relic_wiki.get(str(relic["id"]), "")

            for prop in char["properties"]:
                prop_type = prop["property_type"]
                prop["recommended"] = prop_type in char_rec_props
                prop["preferred"] = prop_type in char_custom_props
                prop["info"] = props_info[str(prop_type)]

            if char["equip"]:
                char["equip"]["wiki"] = equip_wiki.get(str(char["equip"]["id"]), "")

        return values


class StarRailDetailCharacterResponse(StarRailSimpleCharacterResponse):
    """HSR characters endpoint response model for when viewed by the user."""

    avatar_list: Sequence[character.StarRailDetailCharacter]



"""Starrail chronicle stats."""

import typing

from genshin.models.model import Aliased, APIModel

from .. import character

__all__ = [
    "PartialStarRailUserStats",
    "StarRailStats",
    "StarRailUserInfo",
    "StarRailUserStats",
]


class StarRailStats(APIModel):
    """Overall user stats."""

    active_days: int
    avatar_num: int
    achievement_num: int
    chest_num: int
    abyss_process: str
    dreamscape_pass_sticker: int = Aliased("dream_paster_num")


class PartialStarRailUserStats(APIModel):
    """User stats with characters without equipment."""

    stats: StarRailStats
    characters: typing.Sequence[character.StarRailPartialCharacter] = Aliased("avatar_list")
    in_game_avatar: str = Aliased("cur_head_icon_url")
    phone_background: str = Aliased("phone_background_image_url")


class StarRailUserInfo(APIModel):
    """User info."""

    nickname: str
    server: str = Aliased("region")
    level: int
    avatar: str


class StarRailUserStats(PartialStarRailUserStats):
    """User stats."""

    info: StarRailUserInfo



"""Auth endpoints responses models."""

import typing

import pydantic

__all__ = ["Account", "ShieldLoginResponse"]


class Account(pydantic.BaseModel):
    """Account data returned by the shield login endpoint."""

    uid: str
    name: str
    email: str
    mobile: str
    is_email_verify: str
    realname: str
    identity_card: str
    token: str
    safe_mobile: str
    facebook_name: str
    google_name: str
    twitter_name: str
    game_center_name: str
    apple_name: str
    sony_name: str
    tap_name: str
    country: str
    reactivate_ticket: str
    area_code: str
    device_grant_ticket: str
    steam_name: str
    unmasked_email: str
    unmasked_email_type: int
    cx_name: typing.Optional[str] = None


class ShieldLoginResponse(pydantic.BaseModel):
    """Response model for the shield login endpoint."""

    account: Account
    device_grant_required: bool
    safe_moblie_required: bool
    realperson_required: bool
    reactivate_required: bool
    realname_operation: str



"""ZZZ sticky notes (real-time notes) models."""

import datetime
import enum
import logging
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, prevent_enum_error

__all__ = ("BatteryCharge", "VideoStoreState", "ZZZEngagement", "ZZZMemberCard", "ZZZNotes", "ZZZTempleRunning")

logger: logging.Logger = logging.getLogger(__name__)


class VideoStoreState(enum.Enum):
    """Video store management state."""

    REVENUE_AVAILABLE = "SaleStateDone"
    WAITING_TO_OPEN = "SaleStateNo"
    CURRENTLY_OPEN = "SaleStateDoing"


class BenchState(enum.Enum):
    """Crafting state."""

    CAN_PRODUCE = "BenchStateCanProduce"
    """Craftable."""
    PRODUCING = "BenchStateProducing"
    UNKNOWN = "BenchStateUnknown"


class ShelveStoreState(enum.Enum):
    """Selling state."""

    CAN_SELL = "ShelveStateCanSell"
    """Can Be Stocked."""
    SELLING = "ShelveStateSelling"
    SOLD_OUT = "ShelveStateSoldOut"
    """Out of Stock."""
    UNKNOWN = "ShelveStateUnknown"


class ExpeditionState(enum.Enum):
    """Bangboo expedition state."""

    CAN_SEND = "ExpeditionStateInCanSend"
    """Dispatchable."""
    IN_PROGRESS = "ExpeditionStateInProgress"
    """Adventuring."""
    ENDED = "ExpeditionStateEnd"
    """Squad Return."""
    UNKNOWN = "ExpeditionStateUnknown"


class ZZZMemberCardState(enum.Enum):
    """ZZZ Member Card state."""

    MEMBER = "MemberCardStateACK"
    NO = "MemberCardStateNo"


class ZZZCardSignState(enum.Enum):
    """ZZZ Member Card Sign state."""

    NO = "CardSignNo"
    DONE = "CardSignDone"


class BatteryCharge(APIModel):
    """ZZZ battery charge model."""

    current: int
    max: int
    seconds_till_full: int = Aliased("restore")

    @property
    def is_full(self) -> bool:
        """Check if the energy is full."""
        return self.current == self.max

    @property
    def full_datetime(self) -> datetime.datetime:
        """Get the datetime when the energy will be full."""
        return datetime.datetime.now().astimezone() + datetime.timedelta(seconds=self.seconds_till_full)

    @pydantic.model_validator(mode="before")
    @classmethod
    def __unnest_progress(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        return {**values, **values.pop("progress", {})}


class ZZZEngagement(APIModel):
    """ZZZ engagement model."""

    current: int
    max: int


class ZZZAutoWork(APIModel):
    """Temple auto work status."""

    is_auto_work_running: bool
    auto_work_ended: bool
    left_ts: datetime.timedelta


class ZZZTempleRunning(APIModel):
    """ZZZ Suibian Temple Management model."""

    bench_state: typing.Union[BenchState, str]
    currency_next_refresh_ts: datetime.timedelta
    current_currency: int
    expedition_state: typing.Union[ExpeditionState, str]
    level: int
    shelve_state: typing.Union[ShelveStoreState, str]
    weekly_currency_max: int
    auto_work: typing.Optional[ZZZAutoWork] = None

    @property
    def reset_datetime(self) -> datetime.datetime:
        """The datetime when the currency will be reset."""
        return datetime.datetime.now().astimezone() + self.currency_next_refresh_ts

    @pydantic.field_validator("currency_next_refresh_ts", mode="before")
    @classmethod
    def __parse_currency_refresh(cls, v: str) -> datetime.timedelta:
        return datetime.timedelta(seconds=int(v))

    @pydantic.field_validator("bench_state", mode="before")
    @classmethod
    def __parse_bench_state(cls, v: str) -> typing.Union[BenchState, str]:
        return prevent_enum_error(v, BenchState)

    @pydantic.field_validator("expedition_state", mode="before")
    @classmethod
    def __parse_expedition_state(cls, v: str) -> typing.Union[ExpeditionState, str]:
        return prevent_enum_error(v, ExpeditionState)

    @pydantic.field_validator("shelve_state", mode="before")
    @classmethod
    def __parse_shelve_store_state(cls, v: str) -> typing.Union[ShelveStoreState, str]:
        return prevent_enum_error(v, ShelveStoreState)


class ZZZMemberCard(APIModel):
    """ZZZ Member Card model."""

    exp_time: datetime.timedelta
    is_open: bool
    member_card_state: ZZZMemberCardState

    @property
    def reset_datetime(self) -> datetime.datetime:
        """The datetime when the member card will be reset."""
        return datetime.datetime.now().astimezone() + self.exp_time

    @pydantic.field_validator("exp_time", mode="before")
    @classmethod
    def __parse_currency_refresh(cls, v: str) -> datetime.timedelta:
        return datetime.timedelta(seconds=int(v))


class BountyCommission(APIModel):
    """Bounty commission progress model."""

    cur_completed: int = Aliased("num")
    total: int
    refresh_time: datetime.timedelta
    unlock: typing.Optional[bool] = None

    @property
    def completed(self) -> bool:
        """Whether the bounty commissions are completed."""
        return self.cur_completed == self.total

    @property
    def reset_datetime(self) -> datetime.datetime:
        """The datetime when the bounty commission will be reset."""
        return datetime.datetime.now().astimezone() + self.refresh_time


class SurveyPoints(APIModel):
    """ZZZ sticky notes, Hollow Zero investigation points model."""

    num: int
    total: int
    is_max_level: bool  # Not sure what this means


class HollowZero(APIModel):
    """Hollow Zero in ZZZ sticky notes model."""

    bounty_commission: typing.Optional[BountyCommission] = None
    investigation_point: typing.Optional[SurveyPoints] = Aliased("survey_points", default=None)


class WeeklyTask(APIModel):
    """ZZZ sticky notes, Weekly task model."""

    cur_point: int
    max_point: int
    refresh_time: datetime.timedelta

    @property
    def completed(self) -> bool:
        """Whether the weekly tasks is full completed."""
        return self.cur_point >= self.max_point

    @property
    def reset_datetime(self) -> datetime.datetime:
        """The datetime when the weekly tasks will be reset."""
        return datetime.datetime.now().astimezone() + self.refresh_time


class ZZZNotes(APIModel):
    """Zenless Zone Zero sticky notes model."""

    battery_charge: BatteryCharge = Aliased("energy")
    engagement: ZZZEngagement = Aliased("vitality")
    scratch_card_completed: bool = Aliased("card_sign")
    video_store_state: VideoStoreState
    hollow_zero: HollowZero
    weekly_task: typing.Optional[WeeklyTask] = None
    card_sign: ZZZCardSignState
    member_card: ZZZMemberCard
    temple_running: ZZZTempleRunning

    @pydantic.field_validator("scratch_card_completed", mode="before")
    @classmethod
    def __transform_value(cls, v: typing.Literal["CardSignDone", "CardSignNotDone"]) -> bool:
        return v == "CardSignDone"

    @pydantic.model_validator(mode="before")
    @classmethod
    def __unnest_value(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if "video_store_state" not in values:
            values["video_store_state"] = values["vhs_sale"]["sale_state"]
        if "hollow_zero" not in values:
            values["hollow_zero"] = {
                "bounty_commission": values["bounty_commission"],
                "survey_points": values["survey_points"],
            }
        return values



"""Cookie managers for making authenticated requests."""

from __future__ import annotations

import abc
import functools
import http.cookies
import logging
import typing
import warnings

import aiohttp
import aiohttp.typedefs
import yarl

from genshin import errors, types
from genshin.client import ratelimit
from genshin.utility import fs as fs_utility

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "BaseCookieManager",
    "CookieManager",
    "InternationalCookieManager",
    "RotatingCookieManager",
    "parse_cookie",
]

CookieOrHeader = typing.Union["http.cookies.BaseCookie[typing.Any]", typing.Mapping[typing.Any, typing.Any], str]
AnyCookieOrHeader = typing.Union[CookieOrHeader, typing.Sequence[CookieOrHeader]]

T = typing.TypeVar("T")
CallableT = typing.TypeVar("CallableT", bound="typing.Callable[..., object]")
AsyncCallableT = typing.TypeVar("AsyncCallableT", bound="typing.Callable[..., typing.Awaitable[object]]")
MaybeSequence = typing.Union[T, typing.Sequence[T]]


def parse_cookie(cookie: typing.Optional[CookieOrHeader]) -> dict[str, str]:
    """Parse a cookie or header into a cookie mapping."""
    if cookie is None:
        return {}

    if isinstance(cookie, str):
        cookie = http.cookies.SimpleCookie(cookie)

    return {str(k): v.value if isinstance(v, http.cookies.Morsel) else str(v) for k, v in cookie.items()}


def get_cookie_identifier(cookie: typing.Mapping[str, str]) -> typing.Optional[str]:
    """Get a unique identifier for a cookie."""
    for name, value in cookie.items():
        if name in ("ltuid", "account_id", "ltuid_v2", "account_id_v2"):
            return value

    # fallback non-digit identifier
    for name, value in cookie.items():
        if name in ("ltmid_v2", "account_mid_v2"):
            return value

    return None


class BaseCookieManager(abc.ABC):
    """A cookie manager for making requests."""

    _proxy: typing.Optional[yarl.URL] = None
    _socks_proxy: typing.Optional[str] = None

    @classmethod
    def from_cookies(cls, cookies: typing.Optional[AnyCookieOrHeader] = None) -> BaseCookieManager:
        """Create an arbitrary cookie manager implementation instance."""
        if not cookies:
            return CookieManager()

        if isinstance(cookies, typing.Sequence) and not isinstance(cookies, str):
            return RotatingCookieManager(cookies)

        return CookieManager(cookies)

    @classmethod
    def from_browser_cookies(cls, browser: typing.Optional[str] = None) -> CookieManager:
        """Create a cookie manager with browser cookies."""
        manager = CookieManager()
        manager.set_browser_cookies(browser)

        return manager

    @property
    def available(self) -> bool:
        """Whether the authentication cookies are available."""
        return True

    @property
    def multi(self) -> bool:
        """Whether the cookie manager contains multiple cookies and therefore should not cache private data."""
        return False

    @property
    def user_id(self) -> typing.Optional[int]:
        """The id of the user that owns cookies.

        Returns None if not found or not applicable.
        """
        return None

    @property
    def proxy(self) -> typing.Optional[yarl.URL]:
        """Proxy for http(s) requests."""
        return self._proxy

    @proxy.setter
    def proxy(self, proxy: typing.Optional[aiohttp.typedefs.StrOrURL]) -> None:
        if proxy is None:
            self._proxy = None
            self._socks_proxy = None
            return

        proxy = yarl.URL(proxy)

        if proxy.scheme in {"socks4", "socks5"}:
            self._socks_proxy = str(proxy)
            return

        if proxy.scheme not in {"https", "http", "ws", "wss"}:
            raise ValueError("Proxy URL must have a valid scheme.")

        self._proxy = proxy

    def create_session(self, **kwargs: typing.Any) -> aiohttp.ClientSession:
        """Create a client session."""
        if self._socks_proxy is not None:
            import aiohttp_socks

            connector = aiohttp_socks.ProxyConnector.from_url(self._socks_proxy)
        else:
            connector = None

        return aiohttp.ClientSession(
            cookie_jar=aiohttp.DummyCookieJar(),
            connector=connector,
            **kwargs,
        )

    @ratelimit.handle_ratelimits()
    @ratelimit.handle_request_timeouts()
    async def _request(
        self,
        method: str,
        str_or_url: aiohttp.typedefs.StrOrURL,
        cookies: typing.MutableMapping[str, str],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Make a request towards any json resource."""
        async with self.create_session() as session:
            async with session.request(method, str_or_url, proxy=self.proxy, cookies=cookies, **kwargs) as response:
                if response.content_type != "application/json":
                    content = await response.text()
                    raise errors.GenshinException(msg="Recieved a response with an invalid content type:\n" + content)

                data = await response.json()

                if not self.multi:
                    new_cookies = parse_cookie(response.cookies)
                    new_keys = new_cookies.keys() - cookies.keys()
                    if new_keys:
                        cookies.update(new_cookies)
                        _LOGGER.debug("Updating cookies for %s: %s", get_cookie_identifier(cookies), new_keys)

        errors.check_for_geetest(data)

        retcode = data.get("retcode")
        if retcode is None or retcode == 0:
            if "data" in data:
                return data["data"]
            return data

        errors.raise_for_retcode(data)

    @abc.abstractmethod
    async def request(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        method: str = "GET",
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        json: typing.Any = None,
        cookies: typing.Optional[aiohttp.typedefs.LooseCookies] = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Make an authenticated request."""


class CookieManager(BaseCookieManager):
    """Standard implementation of the cookie manager."""

    _cookies: dict[str, str]

    def __init__(
        self,
        cookies: typing.Optional[CookieOrHeader] = None,
    ) -> None:
        self.cookies = parse_cookie(cookies)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.cookies})"

    @property
    def cookies(self) -> typing.MutableMapping[str, str]:
        """Cookies used for authentication."""
        return self._cookies

    @cookies.setter
    def cookies(self, cookies: typing.Optional[CookieOrHeader]) -> None:
        if not cookies:
            self._cookies = {}
            return

        self._cookies = parse_cookie(cookies)

    @property
    def available(self) -> bool:
        return bool(self._cookies)

    @property
    def multi(self) -> bool:
        return False

    @property
    def jar(self) -> http.cookies.SimpleCookie:
        """SimpleCookie containing the cookies."""
        return http.cookies.SimpleCookie(self.cookies)

    @property
    def header(self) -> str:
        """Header representation of cookies.

        This representation is reparsable by the manager.
        """
        return self.jar.output(header="", sep=";").strip()

    def set_cookies(
        self,
        cookies: typing.Optional[CookieOrHeader] = None,
        **kwargs: typing.Any,
    ) -> typing.MutableMapping[str, str]:
        """Parse and set cookies."""
        if not bool(cookies) ^ bool(kwargs):
            raise TypeError("Cannot use both positional and keyword arguments at once")

        self.cookies = parse_cookie(cookies or kwargs)
        return self.cookies

    def set_browser_cookies(self, browser: typing.Optional[str] = None) -> typing.Mapping[str, str]:
        """Extract cookies from your browser and set them as client cookies.

        Available browsers: chrome, chromium, opera, edge, firefox.
        """
        self.cookies = parse_cookie(fs_utility.get_browser_cookies(browser))
        return self.cookies

    @property
    def user_id(self) -> typing.Optional[int]:
        """The id of the user that owns cookies.

        Returns None if cookies are not set.
        """
        for name, value in self.cookies.items():
            if name in ("ltuid", "account_id", "ltuid_v2", "account_id_v2"):
                if not value:
                    raise ValueError(f"{name} can not be an empty string.")

                return int(value)

        return None

    async def request(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        method: str = "GET",
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Make an authenticated request."""
        return await self._request(method, url, cookies=self.cookies, **kwargs)


class CookieSequence(typing.Sequence[typing.Mapping[str, str]]):
    MAX_USES: int = 30

    # {id: ({cookie}, uses), ...}
    _cookies: dict[str, tuple[dict[str, str], int]]

    def __init__(self, cookies: typing.Optional[typing.Sequence[CookieOrHeader]] = None) -> None:
        self.cookies = [parse_cookie(cookie) for cookie in cookies or []]

    @property
    def cookies(self) -> typing.Sequence[typing.Mapping[str, str]]:
        """Cookies used for authentication"""
        cookies = sorted(self._cookies.values(), key=lambda x: 0 if x[1] >= self.MAX_USES else x[1], reverse=True)
        return [cookies for cookies, _ in cookies]

    @cookies.setter
    def cookies(self, cookies: typing.Optional[typing.Sequence[CookieOrHeader]]) -> None:
        if not cookies:
            self._cookies = {}
            return

        self._cookies = {}
        for cookie in cookies:
            cookie = parse_cookie(cookie)

            account_id = get_cookie_identifier(cookie)
            if not account_id or not account_id.isdigit():
                raise ValueError(f"Cookies must contain a valid identifier: {cookie}")

            if account_id in self._cookies:
                raise ValueError(f"Cannot use the same identifier for multiple cookies: {account_id}.")

            self._cookies[account_id] = (cookie, 0)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} len={len(self._cookies)}>"

    def __getitem__(self, index: int) -> typing.Mapping[str, str]:  # type: ignore # I can't be fucked with slices
        return self.cookies[index]

    def __len__(self) -> int:
        return len(self.cookies)

    def __iter__(self) -> typing.Iterator[typing.Mapping[str, str]]:
        return iter(self.cookies)


class RotatingCookieManager(BaseCookieManager):
    """Cookie Manager with rotating cookies."""

    _cookies: CookieSequence

    def __init__(self, cookies: typing.Optional[typing.Sequence[CookieOrHeader]] = None) -> None:
        self.set_cookies(cookies)

    @property
    def cookies(self) -> typing.Sequence[typing.Mapping[str, str]]:
        """Cookies used for authentication"""
        return self._cookies

    @cookies.setter
    def cookies(self, cookies: typing.Optional[typing.Sequence[CookieOrHeader]]) -> None:
        self._cookies.cookies = cookies  # type: ignore # mypy does not understand property setters

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} len={len(self._cookies)}>"

    @property
    def available(self) -> bool:
        return bool(self._cookies)

    @property
    def multi(self) -> bool:
        return True

    def set_cookies(
        self,
        cookies: typing.Optional[typing.Sequence[CookieOrHeader]] = None,
    ) -> typing.Sequence[typing.Mapping[str, str]]:
        """Parse and set cookies."""
        self._cookies = CookieSequence(cookies)
        return self.cookies

    async def request(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        method: str = "GET",
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Make an authenticated request."""
        if not self.cookies:
            raise RuntimeError("Tried to make a request before setting cookies")

        for account_id, (cookie, uses) in self._cookies._cookies.copy().items():
            try:
                data = await self._request(method, url, cookies=cookie, **kwargs)
            except errors.TooManyRequests:
                _LOGGER.debug("Putting cookie %s on cooldown.", account_id)
                self._cookies._cookies[account_id] = (cookie, self._cookies.MAX_USES)
            except errors.InvalidCookies:
                warnings.warn(f"Deleting invalid cookie {cookie}")
                # prevent race conditions
                if account_id in self._cookies._cookies:
                    del self._cookies._cookies[account_id]
            else:
                self._cookies._cookies[account_id] = (cookie, 1 if uses >= self._cookies.MAX_USES else uses + 1)
                return data

        msg = "All cookies have hit their request limit of 30 accounts per day."
        raise errors.TooManyRequests({"retcode": 10101}, msg)


class InternationalCookieManager(BaseCookieManager):
    """Cookie Manager with international rotating cookies."""

    _cookies: typing.Mapping[types.Region, CookieSequence]

    def __init__(self, cookies: typing.Optional[typing.Mapping[str, MaybeSequence[CookieOrHeader]]] = None) -> None:
        self.set_cookies(cookies)

    @property
    def cookies(self) -> typing.Mapping[types.Region, typing.Sequence[typing.Mapping[str, str]]]:
        """Cookies used for authentication"""
        return self._cookies

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def available(self) -> bool:
        return bool(self._cookies)

    @property
    def multi(self) -> bool:
        return True

    def set_cookies(
        self,
        cookies: typing.Optional[typing.Mapping[str, MaybeSequence[CookieOrHeader]]] = None,
    ) -> typing.Mapping[types.Region, typing.Sequence[typing.Mapping[str, str]]]:
        """Parse and set cookies."""
        self._cookies = {}
        if not cookies:
            return {}

        for region, regional_cookies in cookies.items():
            if not isinstance(regional_cookies, typing.Sequence):
                regional_cookies = [regional_cookies]

            self._cookies[types.Region(region)] = CookieSequence(regional_cookies)

        return self.cookies

    def guess_region(self, url: yarl.URL) -> types.Region:
        """Guess the region from the URL."""
        assert url.host is not None

        if "os" in url.host or "os" in url.path:
            return types.Region.OVERSEAS

        if "takumi" in url.host:
            return types.Region.CHINESE

        if "sg" in url.host:
            return types.Region.OVERSEAS

        return types.Region.CHINESE

    async def request(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        method: str = "GET",
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Make an authenticated request."""
        if not self.cookies:
            raise RuntimeError("Tried to make a request before setting cookies")

        region = self.guess_region(yarl.URL(url))

        # TODO: less copy-paste
        for account_id, (cookie, uses) in self._cookies[region]._cookies.copy().items():
            try:
                data = await self._request(method, url, cookies=cookie, **kwargs)
            except errors.TooManyRequests:
                _LOGGER.debug("Putting cookie %s on cooldown.", account_id)
                self._cookies[region]._cookies[account_id] = (cookie, self._cookies[region].MAX_USES)
            except errors.InvalidCookies:
                warnings.warn(f"Deleting invalid cookie {cookie}")
                # prevent race conditions
                if account_id in self._cookies[region]._cookies:
                    del self._cookies[region]._cookies[account_id]
            else:
                self._cookies[region]._cookies[account_id] = (
                    cookie,
                    1 if uses >= self._cookies[region].MAX_USES else uses + 1,
                )
                return data

        msg = "All cookies have hit their request limit of 30 accounts per day."
        raise errors.TooManyRequests({"retcode": 10101}, msg)


def no_multi(func: CallableT) -> CallableT:
    """Prevent function to be ran with a multi-cookie manager."""

    @functools.wraps(func)
    def wrapper(self: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if not hasattr(self, "cookie_manager"):
            raise TypeError("Cannot use @no_multi on a plain function.")
        if self.cookie_manager.multi:
            raise RuntimeError(f"Cannot use {func.__name__} with multi-cookie managers - data is private.")

        return func(self, *args, **kwargs)

    return typing.cast("CallableT", wrapper)


def requires_cookie_token(func: AsyncCallableT) -> AsyncCallableT:
    """Prevent function to be ran without a cookie_token."""

    @functools.wraps(func)
    async def wrapper(self: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if not hasattr(self, "cookie_manager"):
            raise TypeError("Cannot use @requires_cookie_token on a plain function.")
        if self.cookie_manager.multi:
            raise RuntimeError(f"Cannot use {func.__name__} with multi-cookie managers - data is private.")
        if isinstance(self.cookie_manager, CookieManager):
            contained = set(self.cookie_manager.cookies.keys())
            if "cookie_token" not in contained and "cookie_token_v2" not in contained:
                raise errors.InvalidCookies(msg="Missing cookie_token or cookie_token_v2 in cookies.")

        return await func(self, *args, **kwargs)

    return typing.cast("AsyncCallableT", wrapper)



import datetime
import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime, TZDateTime
from genshin.models.starrail.chronicle.base import PartialTime
from genshin.models.zzz.character import ZZZElementType, ZZZSpecialty

__all__ = (
    "ChallengeBangboo",
    "DeadlyAssault",
    "DeadlyAssaultAgent",
    "DeadlyAssaultBoss",
    "DeadlyAssaultBuff",
    "DeadlyAssaultChallenge",
    "ShiyuDefense",
    "ShiyuDefenseBangboo",
    "ShiyuDefenseBuff",
    "ShiyuDefenseCharacter",
    "ShiyuDefenseFloor",
    "ShiyuDefenseMonster",
    "ShiyuDefenseNode",
    "ShiyuDefenseV1",
    "ShiyuDefenseV2",
    "ShiyuMonsterElementEffect",
    "ShiyuMonsterElementEffects",
    "ShiyuV2FifthFloor",
    "ShiyuV2FifthFloorLayer",
    "ShiyuV2FourthFloor",
    "ShiyuV2FourthFloorLayer",
    "ThresholdSimulation",
    "ThresholdSimulationBangboo",
    "ThresholdSimulationBoss",
    "ThresholdSimulationBossChallenge",
    "ThresholdSimulationBuff",
    "ThresholdSimulationChallenge",
    "ThresholdSimulationCharacter",
    "ThresholdSimulationInfo",
    "ThresholdSimulationMainChallenge",
    "ThresholdSimulationPlayer",
)


class ShiyuMonsterElementEffect(enum.IntEnum):
    """Shiyu Defense monster element effect enum."""

    WEAKNESS = 1
    RESISTANCE = -1
    NEUTRAL = 0


class ShiyuDefenseBangboo(APIModel):
    """Shiyu Defense bangboo model."""

    id: int
    rarity: typing.Literal["S", "A"]
    level: int
    icon: str = Aliased("bangboo_rectangle_url")


class ChallengeBangboo(ShiyuDefenseBangboo):
    """Bangboo model for backward compatibility."""


class ShiyuDefenseCharacter(APIModel):
    """Shiyu Defense character model."""

    id: int
    level: int
    rarity: typing.Literal["S", "A"]
    element: ZZZElementType = Aliased("element_type")
    icon: str = Aliased("role_square_url")
    mindscape: int = Aliased("rank")


class ShiyuDefenseBuff(APIModel):
    """Shiyu Defense buff model."""

    name: str = Aliased("title")
    description: str = Aliased("text")


class ShiyuMonsterElementEffects(pydantic.BaseModel):
    """Shiyu Defense monster element effects model."""

    ice: ShiyuMonsterElementEffect = Aliased("ice_weakness")
    fire: ShiyuMonsterElementEffect = Aliased("fire_weakness")
    electric: ShiyuMonsterElementEffect = Aliased("elec_weakness")
    ether: ShiyuMonsterElementEffect = Aliased("ether_weakness")
    physical: ShiyuMonsterElementEffect = Aliased("physics_weakness")


class ShiyuDefenseMonster(APIModel):
    """Shiyu Defense monster model."""

    id: int
    name: str
    weakness: typing.Union[ZZZElementType, int] = Aliased(
        "weak_element_type", deprecated="ShiyuDefenseMonster.weakness doesn't return anything meaningful anymore."
    )
    level: int
    element_effects: ShiyuMonsterElementEffects

    @pydantic.model_validator(mode="before")
    @classmethod
    def __nest_element_effects(cls, v: dict[str, typing.Any]) -> dict[str, typing.Any]:
        v["element_effects"] = {
            "ice_weakness": v["ice_weakness"],
            "fire_weakness": v["fire_weakness"],
            "elec_weakness": v["elec_weakness"],
            "ether_weakness": v["ether_weakness"],
            "physics_weakness": v["physics_weakness"],
        }
        return v


class ShiyuDefenseNode(APIModel):
    """Shiyu Defense node model."""

    characters: list[ShiyuDefenseCharacter] = Aliased("avatars")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)
    recommended_elements: list[ZZZElementType] = Aliased("element_type_list")
    enemies: list[ShiyuDefenseMonster] = Aliased("monster_info")
    battle_time: typing.Optional[datetime.timedelta] = None

    @pydantic.field_validator("enemies", mode="before")
    @classmethod
    def __convert_enemies(cls, value: dict[typing.Literal["level", "list"], typing.Any]) -> list[ShiyuDefenseMonster]:
        level = value["level"]
        result: list[ShiyuDefenseMonster] = []
        for monster in value["list"]:
            monster["level"] = level
            result.append(ShiyuDefenseMonster(**monster))
        return result


class ShiyuDefenseFloor(APIModel):
    """Shiyu Defense floor model."""

    index: int = Aliased("layer_index")
    rating: typing.Literal["S", "A", "B"]
    id: int = Aliased("layer_id")
    buffs: list[ShiyuDefenseBuff]
    node_1: ShiyuDefenseNode
    node_2: ShiyuDefenseNode
    challenge_time: TZDateTime = Aliased("floor_challenge_time")
    name: str = Aliased("zone_name")

    @pydantic.field_validator("challenge_time", mode="before")
    @classmethod
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> typing.Optional[TZDateTime]:
        if value:
            return datetime.datetime(**value)
        return None


class ShiyuDefenseV1(APIModel):
    """ZZZ Shiyu Defense V1 model."""

    schedule_id: int
    begin_time: typing.Optional[DateTime] = Aliased("hadal_begin_time")
    end_time: typing.Optional[DateTime] = Aliased("hadal_end_time")
    has_data: bool
    ratings: typing.Mapping[typing.Literal["S", "A", "B"], int] = Aliased("rating_list")
    floors: list[ShiyuDefenseFloor] = Aliased("all_floor_detail")
    fastest_clear_time: int = Aliased("fast_layer_time")
    """Fastest clear time this season in seconds."""
    max_floor: int = Aliased("max_layer")

    @pydantic.field_validator("ratings", mode="before")
    @classmethod
    def __convert_ratings(
        cls, v: list[dict[typing.Literal["times", "rating"], typing.Any]]
    ) -> typing.Mapping[typing.Literal["S", "A", "B"], int]:
        return {d["rating"]: d["times"] for d in v}

    @pydantic.computed_field  # type: ignore[prop-decorator]
    @property
    def total_clear_time(self) -> int:
        """Total clear time for all floors in seconds."""
        total = 0
        for floor in self.floors:
            for node in (floor.node_1, floor.node_2):
                if node.battle_time is None:
                    continue
                total += int(node.battle_time.total_seconds())
        return total


ShiyuDefense = ShiyuDefenseV1  # Backward compatibility


class ShiyuV2FifthFloorLayer(APIModel):
    """ZZZ Shiyu Defense V2 fifth floor layer model."""

    id: int = Aliased("layer_id")
    rating: typing.Literal["S", "A", "B"]
    buff: ShiyuDefenseBuff = Aliased("buffer")
    score: int
    max_score: int
    clear_time: int = Aliased("battle_time")
    """Clear time in seconds."""
    boss_icon: str = Aliased("monster_pic")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)
    characters: typing.Sequence[ShiyuDefenseCharacter] = Aliased("avatar_list")


class ShiyuV2FifthFloor(APIModel):
    """ZZZ Shiyu Defense V2 fifth floor model."""

    layers: typing.Sequence[ShiyuV2FifthFloorLayer] = Aliased("layer_challenge_info_list")


class ShiyuV2FourthFloorLayer(APIModel):
    """ZZZ Shiyu Defense V2 fourth floor layer model."""

    id: int = Aliased("layer_id")
    clear_time: int = Aliased("battle_time")
    """Clear time in seconds."""
    characters: typing.Sequence[ShiyuDefenseCharacter] = Aliased("avatar_list")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)


class ShiyuV2FourthFloor(APIModel):
    """ZZZ Shiyu Defense V2 fourth floor model."""

    buff: ShiyuDefenseBuff = Aliased("buffer")
    challenge_time: DateTime
    rating: typing.Literal["S", "A", "B"]
    layers: typing.Sequence[ShiyuV2FourthFloorLayer] = Aliased("layer_challenge_info_list")


class ShiyuV2BriefInfo(APIModel):
    """ZZZ Shiyu Defense V2 brief info."""

    score: int
    max_score: int
    rank_percent: str
    total_clear_time: int = Aliased("battle_time")
    rating: typing.Literal["S+", "S", "A", "B"]
    challenge_time: DateTime

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"


class ShiyuDefenseV2(APIModel):
    """ZZZ Shiyu Defense V2 model."""

    schedule_id: int = Aliased("zone_id")
    begin_time: typing.Optional[DateTime] = Aliased("hadal_begin_time")
    end_time: typing.Optional[DateTime] = Aliased("hadal_end_time")
    passed_fifth_floor: bool = Aliased("pass_fifth_floor")

    brief_info: ShiyuV2BriefInfo = Aliased("brief")
    fourth_frontier: ShiyuV2FourthFloor = Aliased("fourth_layer_detail")
    fifth_frontier: ShiyuV2FifthFloor = Aliased("fitfh_layer_detail")  # Nice typo hoyo

    player_nickname: str = Aliased("nick_name")
    player_avatar: str = Aliased("icon")


class DeadlyAssaultBoss(APIModel):
    """ZZZ Deadly Assault boss."""

    icon: str
    name: str
    background: str = Aliased("bg_icon")
    badge_icon: str = Aliased("race_icon")


class DeadlyAssaultBuff(APIModel):
    """ZZZ Deadly Assault buff model."""

    name: str
    description: str = Aliased("desc")
    icon: str


class DeadlyAssaultAgent(APIModel):
    """ZZZ Deadly Assault agent model."""

    id: int
    level: int
    element: ZZZElementType = Aliased("element_type")
    specialty: ZZZSpecialty = Aliased("avatar_profession")
    rarity: typing.Literal["S", "A"]
    mindscape: int = Aliased("rank")
    icon: str = Aliased("role_square_url")


class DeadlyAssaultChallenge(APIModel):
    """ZZZ Deadly Assault challenge model."""

    score: int
    star: int
    total_star: int
    challenge_time: datetime.datetime

    boss: DeadlyAssaultBoss
    buffs: typing.Sequence[DeadlyAssaultBuff] = Aliased("buffer")
    agents: typing.Sequence[DeadlyAssaultAgent] = Aliased("avatar_list")
    bangboo: typing.Optional[ShiyuDefenseBangboo] = Aliased("buddy", default=None)

    @pydantic.field_validator("challenge_time", mode="before")
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> typing.Optional[TZDateTime]:
        if value:
            return datetime.datetime(**value)
        return None

    @pydantic.field_validator("boss", mode="before")
    def __parse_boss(cls, value: typing.List[typing.Mapping[str, typing.Any]]) -> DeadlyAssaultBoss:
        if not value:
            raise ValueError("No boss data provided.")
        return DeadlyAssaultBoss(**value[0])


class DeadlyAssault(APIModel):
    """ZZZ Deadly Assault model."""

    id: int = Aliased("zone_id")
    start_time: typing.Optional[DateTime]
    end_time: typing.Optional[DateTime]

    challenges: typing.Sequence[DeadlyAssaultChallenge] = Aliased("list")
    has_data: bool
    total_score: int
    total_star: int
    rank_percent: str

    nickname: str = Aliased("nick_name")
    player_avatar: str = Aliased("avatar_icon")

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"


class ThresholdSimulationInfo(APIModel):
    """ZZZ Threshold Simulation brief info."""

    id: int = Aliased("void_front_id")
    time_remaining_over_42_days: bool = Aliased("end_ts_over_42_days")
    time_remaining_days: int = Aliased("end_ts")
    has_data: bool = Aliased("has_ending_record")
    ending_name: str = Aliased("ending_record_name")
    ending_background: str = Aliased("ending_record_bg_pic")
    total_score: int
    rank_percent: str

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"


class ThresholdSimulationCharacter(APIModel):
    """ZZZ Threshold Simulation character model."""

    id: int
    level: int
    element: ZZZElementType = Aliased("element_type")
    rarity: typing.Literal["S", "A"]
    mindscape: int = Aliased("rank")
    icon: str = Aliased("role_square_url")


class ThresholdSimulationBangboo(APIModel):
    """ZZZ Threshold Simulation bangboo model."""

    id: int
    rarity: typing.Literal["S", "A"]
    level: int
    icon: str = Aliased("bangboo_rectangle_url")


class ThresholdSimulationBoss(APIModel):
    """ZZZ Threshold Simulation boss model."""

    icon: str
    name: str
    badge_icon: str = Aliased("race_icon")
    background: str = Aliased("bg_icon")


class ThresholdSimulationBuff(APIModel):
    """ZZZ Threshold Simulation buff model."""

    name: str
    description: str = Aliased("desc")
    icon: str


class ThresholdSimulationChallenge(APIModel):
    """ZZZ Threshold Simulation challenge model."""

    id: int = Aliased("battle_id")
    name: str
    rating: typing.Literal["S", "A", "B"] = Aliased("star")

    characters: typing.Sequence[ThresholdSimulationCharacter] = Aliased("avatar_list")
    bangboo: typing.Optional[ThresholdSimulationBangboo] = Aliased("buddy", default=None)
    buff: ThresholdSimulationBuff = Aliased("buffer")


class ThresholdSimulationMainChallenge(ThresholdSimulationChallenge):
    """ZZZ Threshold Simulation main challenge model."""

    node_id: int
    score: int
    max_score: int
    score_multiplier: str = Aliased("score_ratio")
    time: PartialTime = Aliased("challenge_time")

    sub_challenges: typing.Sequence["ThresholdSimulationChallenge"] = Aliased("sub_challenge_record")


class ThresholdSimulationBossChallenge(APIModel):
    """ZZZ Threshold Simulation boss challenge model."""

    boss: ThresholdSimulationBoss = Aliased("boss_info")
    challenge: ThresholdSimulationMainChallenge = Aliased("main_challenge_record")


class ThresholdSimulationPlayer(APIModel):
    """ZZZ Threshold Simulation player info."""

    nickname: str
    server: str
    icon: str


class ThresholdSimulation(APIModel):
    """ZZZ Threshold Simulation."""

    info: ThresholdSimulationInfo = Aliased("void_front_battle_abstract_info_brief")
    boss_challenge: ThresholdSimulationBossChallenge = Aliased("boss_challenge_record")
    challenges: typing.Sequence[ThresholdSimulationChallenge] = Aliased("main_challenge_record_list")
    player: ThresholdSimulationPlayer = Aliased("role_basic_info")



"""Hollow Zero models."""

import datetime
import typing
from enum import IntEnum

from genshin.models.model import Aliased, APIModel

__all__ = (
    "LostVoidCommissionProgress",
    "LostVoidData",
    "LostVoidDataType",
    "LostVoidExplorationLog",
    "LostVoidLicense",
    "LostVoidSummary",
    "LostVoidUltimateChallenge",
)


class LostVoidLicense(APIModel):
    """Lost void license model."""

    cur_level: int
    max_level: int
    icon: str


class LostVoidExplorationLog(APIModel):
    """Lost void exploration log model."""

    current: int = Aliased("cur_task")
    total: int = Aliased("max_task")


class LostVoidCommissionProgress(APIModel):
    """Lost void bounty commission progress model."""

    current: int = Aliased("cur_duty")
    total: int = Aliased("max_duty")


class LostVoidUltimateChallenge(APIModel):
    """Lost void ultimate challenge model."""

    highest_difficulty: str = Aliased("max_name")
    """Highest Difficulty Cleared."""
    total_clears: int = Aliased("max_count")
    """Total Clears on the Highest Difficulty Completed."""
    shortest_clear_time: datetime.timedelta = Aliased("best_time")
    """Shortest Clear Time on the Highest Difficulty Completed."""

    has_data: bool


class LostVoidDataType(IntEnum):
    """Lost void data type enum."""

    EXPLORERS_BADGE = 1
    GEAR_DATABASE = 2
    RESONIUM_RESEARCH = 3
    CARD_DATABASE = 4
    TACTICAL_PRISM_PROGRAM = 5
    UNOFFICIAL_GUIDE = 6
    AUXILLARY_GEAR_DATABASE = 7


class LostVoidData(APIModel):
    """Lost void data model."""

    type: LostVoidDataType
    collected: int = Aliased("cur_collect")
    total: int = Aliased("max_collect")


class LostVoidSummary(APIModel):
    """Lost void summary model."""

    time_remaining: datetime.timedelta = Aliased("refresh_time")
    unlocked: bool = Aliased("unlock")

    license: LostVoidLicense = Aliased("abyss_level")
    exploration_log: LostVoidExplorationLog = Aliased("abyss_task")
    commission_progress: LostVoidCommissionProgress = Aliased("abyss_duty")
    ultimate_challenge: LostVoidUltimateChallenge = Aliased("abyss_max")
    data_collected: typing.Sequence[LostVoidData] = Aliased("abyss_collect")



import typing

import pydantic

from genshin.models.model import TZDateTime

__all__ = ("BattleField", "GodWar", "GreedyEndless", "HonkaiNotes", "UltraEndless")


class HonkaiNotesEvent(pydantic.BaseModel):
    """Base event model."""

    end_time: TZDateTime = pydantic.Field(alias="schedule_end")
    is_open: bool


class GreedyEndless(HonkaiNotesEvent):
    """Greedy Endless event model."""

    current_reward: int = pydantic.Field(alias="cur_reward")
    max_reward: int


class UltraEndless(HonkaiNotesEvent):
    """Ultra Endless event model."""

    group_level: int
    challenge_score: int


class BattleField(HonkaiNotesEvent):
    """Battle Field event model."""

    current_reward: int = pydantic.Field(alias="cur_reward")
    max_reward: int
    current_sss_reward: int = pydantic.Field(alias="cur_sss_reward")
    max_sss_reward: int


class GodWar(HonkaiNotesEvent):
    """God War event model."""

    current_reward: int = pydantic.Field(alias="cur_reward")
    max_reward: int


class HonkaiNotes(pydantic.BaseModel):
    """Honkai Impact 3rd real-time notes model."""

    current_stamina: int
    max_stamina: int
    stamina_recover_time: int
    current_train_score: int

    greedy_endless: GreedyEndless
    ultra_endless: UltraEndless
    battle_field: typing.Optional[BattleField] = None
    god_war: GodWar



"""Base paginators made specifically for interaction with the api."""

from __future__ import annotations

import abc
import typing
import warnings

from . import base

if typing.TYPE_CHECKING:
    from genshin import models


__all__ = ["CursorPaginator", "PagedPaginator", "TokenPaginator"]

T = typing.TypeVar("T")
T_co = typing.TypeVar("T_co", covariant=True)
UniqueT = typing.TypeVar("UniqueT", bound="models.Unique")


class GetterCallback(typing.Protocol[T_co]):
    """Callback for returning resources based on a page or cursor."""

    async def __call__(self, page: int, /) -> typing.Sequence[T_co]:
        """Return a sequence of resources."""
        ...


class TokenGetterCallback(typing.Protocol[T_co]):
    """Callback for returning resources based on a page or cursor."""

    async def __call__(self, token: str, /) -> tuple[str, typing.Sequence[T_co]]:
        """Return a sequence of resources."""
        ...


class APIPaginator(typing.Generic[T], base.BufferedPaginator[T], abc.ABC):
    """Paginator for interaction with the api."""

    __slots__ = ("getter",)

    getter: typing.Callable[..., typing.Awaitable[object]]
    """Underlying getter that yields the next page."""


class PagedPaginator(typing.Generic[T], APIPaginator[T]):
    """Paginator for resources which only require a page number.

    Due to ratelimits the requests must be sequential.
    """

    __slots__ = ("_page_size", "current_page")

    getter: GetterCallback[T]
    """Underlying getter that yields the next page."""

    _page_size: typing.Optional[int]
    """Expected non-zero page size to be able to tell the end."""

    current_page: typing.Optional[int]
    """Current page counter.."""

    def __init__(
        self,
        getter: GetterCallback[T],
        *,
        limit: typing.Optional[int] = None,
        page_size: typing.Optional[int] = None,
    ) -> None:
        super().__init__(limit=limit)
        self.getter = getter
        self._page_size = page_size

        self.current_page = 1

    async def next_page(self) -> typing.Optional[typing.Iterable[T]]:
        """Get the next page of the paginator."""
        if self.current_page is None:
            return None

        data = await self.getter(self.current_page)

        if self._page_size is None:
            warnings.warn("No page size specified for resource, having to guess.")
            self._page_size = len(data)

        if len(data) < self._page_size:
            self.current_page = None
            return data

        self.current_page += 1
        return data


class TokenPaginator(typing.Generic[T], APIPaginator[T]):
    """Paginator for resources which require a token."""

    __slots__ = ("_page_size", "token")

    getter: TokenGetterCallback[T]
    """Underlying getter that yields the next page."""

    _page_size: typing.Optional[int]
    """Expected non-zero page size to be able to tell the end."""

    token: typing.Optional[str]

    def __init__(
        self,
        getter: TokenGetterCallback[T],
        *,
        limit: typing.Optional[int] = None,
        page_size: typing.Optional[int] = None,
    ) -> None:
        super().__init__(limit=limit)
        self.getter = getter
        self._page_size = page_size

        self.token = ""

    async def next_page(self) -> typing.Optional[typing.Iterable[T]]:
        """Get the next page of the paginator."""
        if self.token is None:
            return None

        self.token, data = await self.getter(self.token)

        if self._page_size is None:
            warnings.warn("No page size specified for resource, having to guess.")
            self._page_size = len(data)

        if len(data) < self._page_size:
            self.token = None
            return data

        return data


class CursorPaginator(typing.Generic[UniqueT], APIPaginator[UniqueT]):
    """Paginator based on end_id cursors."""

    __slots__ = ("_page_size", "end_id")

    getter: GetterCallback[UniqueT]
    """Underlying getter that yields the next page."""

    _page_size: typing.Optional[int]
    """Expected non-zero page size to be able to tell the end."""

    end_id: typing.Optional[int]
    """Current end id. If none then exhausted."""

    def __init__(
        self,
        getter: GetterCallback[UniqueT],
        *,
        limit: typing.Optional[int] = None,
        end_id: int = 0,
        page_size: typing.Optional[int] = 20,
    ) -> None:
        super().__init__(limit=limit)
        self.getter = getter
        self.end_id = end_id

        self._page_size = page_size

    async def next_page(self) -> typing.Optional[typing.Iterable[UniqueT]]:
        """Get the next page of the paginator."""
        if self.end_id is None:
            return None

        data = await self.getter(self.end_id)

        if self._page_size is None:
            warnings.warn("No page size specified for resource, having to guess.")
            self._page_size = len(data)

        if len(data) < self._page_size:
            self.end_id = None
            return data

        self.end_id = data[-1].id
        return data



import typing

import pydantic

from genshin.models.model import Aliased, APIModel
from genshin.models.starrail.chronicle.base import PartialTime
from genshin.models.zzz.character import ZZZElementType

__all__ = (
    "ThresholdSimulation",
    "ThresholdSimulationBangboo",
    "ThresholdSimulationBoss",
    "ThresholdSimulationBossChallenge",
    "ThresholdSimulationBuff",
    "ThresholdSimulationChallenge",
    "ThresholdSimulationCharacter",
    "ThresholdSimulationInfo",
    "ThresholdSimulationMainChallenge",
    "ThresholdSimulationPlayer",
)


class ThresholdSimulationInfo(APIModel):
    """ZZZ Threshold Simulation brief info."""

    id: int = Aliased("void_front_id")
    time_remaining_over_42_days: bool = Aliased("end_ts_over_42_days")
    time_remaining_days: int = Aliased("end_ts")
    has_data: bool = Aliased("has_ending_record")
    ending_name: str = Aliased("ending_record_name")
    ending_background: str = Aliased("ending_record_bg_pic")
    total_score: int
    rank_percent: str

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"


class ThresholdSimulationCharacter(APIModel):
    """ZZZ Threshold Simulation character model."""

    id: int
    level: int
    element: ZZZElementType = Aliased("element_type")
    rarity: typing.Literal["S", "A"]
    mindscape: int = Aliased("rank")
    icon: str = Aliased("role_square_url")


class ThresholdSimulationBangboo(APIModel):
    """ZZZ Threshold Simulation bangboo model."""

    id: int
    rarity: typing.Literal["S", "A"]
    level: int
    icon: str = Aliased("bangboo_rectangle_url")


class ThresholdSimulationBoss(APIModel):
    """ZZZ Threshold Simulation boss model."""

    icon: str
    name: str
    badge_icon: str = Aliased("race_icon")
    background: str = Aliased("bg_icon")


class ThresholdSimulationBuff(APIModel):
    """ZZZ Threshold Simulation buff model."""

    name: str
    description: str = Aliased("desc")
    icon: str


class ThresholdSimulationChallenge(APIModel):
    """ZZZ Threshold Simulation challenge model."""

    id: int = Aliased("battle_id")
    name: str
    rating: typing.Literal["S", "A", "B"] = Aliased("star")

    characters: typing.Sequence[ThresholdSimulationCharacter] = Aliased("avatar_list")
    bangboo: typing.Optional[ThresholdSimulationBangboo] = Aliased("buddy", default=None)
    buff: ThresholdSimulationBuff = Aliased("buffer")


class ThresholdSimulationMainChallenge(ThresholdSimulationChallenge):
    """ZZZ Threshold Simulation main challenge model."""

    node_id: int
    score: int
    max_score: int
    score_multiplier: str = Aliased("score_ratio")
    time: PartialTime = Aliased("challenge_time")

    sub_challenges: typing.Sequence["ThresholdSimulationChallenge"] = Aliased("sub_challenge_record")


class ThresholdSimulationBossChallenge(APIModel):
    """ZZZ Threshold Simulation boss challenge model."""

    boss: ThresholdSimulationBoss = Aliased("boss_info")
    challenge: ThresholdSimulationMainChallenge = Aliased("main_challenge_record")


class ThresholdSimulationPlayer(APIModel):
    """ZZZ Threshold Simulation player info."""

    nickname: str
    server: str
    icon: str


class ThresholdSimulation(APIModel):
    """ZZZ Threshold Simulation."""

    info: ThresholdSimulationInfo = Aliased("void_front_battle_abstract_info_brief")
    boss_challenge: ThresholdSimulationBossChallenge = Aliased("boss_challenge_record")
    challenges: typing.Sequence[ThresholdSimulationChallenge] = Aliased("main_challenge_record_list")
    player: ThresholdSimulationPlayer = Aliased("role_basic_info")



import datetime
import enum
import typing

import pydantic

from genshin import types
from genshin.models.model import Aliased, APIModel, TZDateTime, prevent_enum_error

__all__ = (
    "MimoGame",
    "MimoLotteryInfo",
    "MimoLotteryResult",
    "MimoLotteryReward",
    "MimoShopItem",
    "MimoShopItemStatus",
    "MimoTask",
    "MimoTaskStatus",
    "MimoTaskType",
    "PartialMimoLotteryReward",
)


class MimoTaskStatus(enum.IntEnum):
    """Mimo task status."""

    FINISHED = 1
    ONGOING = 2
    CLAIMED = 3


class MimoTaskType(enum.IntEnum):
    """Mimo task type."""

    FINISHABLE = 1
    """e.g. Sunday Advanced Tutorial: What is the core Charge mechanic?"""
    VISIT = 2
    """e.g. Visit the 【Honkai: Star Rail】 Interest Group on the day"""
    COMMENT = 3
    """e.g. Participate in this week's creative interactions and leave your creations in the comments"""
    VIEW_TOPIC = 4
    """e.g. View the "Roaming Through the Realm of Saurians" topic"""
    GI_LOGIN = 5
    """e.g. Log into Genshin Impact today"""
    GI_GAME = 6
    """e.g. Claim rewards from Ley Line Blossoms 2 times"""
    GI_COMMUNITY = 101
    """e.g. Browse today's must-see highlights"""
    HSR_GAME = 8
    """e.g. Complete Divergent Universe or Simulated Universe 1 time"""
    TRAILER = 10
    """e.g. Myriad Celestia Trailer — "After the Sunset" | Honkai: Star Rail"""
    ZZZ_DAILY_LOGIN = 12
    """e.g. Log into Zenless Zone Zero today"""
    ZZZ_CONSECUTIVE_LOGIN = 13
    """e.g. Log into the game for 7 days"""
    ZZZ_GAME = 16
    """e.g. Reach 400 Engagement today"""


class MimoShopItemStatus(enum.IntEnum):
    """Mimo shop item status."""

    EXCHANGEABLE = 1
    NOT_EXCHANGEABLE = 2
    LIMIT_REACHED = 3
    SOLD_OUT = 4


class MimoGame(APIModel):
    """Mimo game."""

    id: int = Aliased("game_id")
    version_id: int
    expire_point: bool
    point: int
    start_time: TZDateTime
    end_time: TZDateTime

    @property
    def game(self) -> typing.Union[typing.Literal["hoyolab"], types.Game, int]:
        if self.id == 5:
            return "hoyolab"
        if self.id == 6:
            return types.Game.STARRAIL
        if self.id == 8:
            return types.Game.ZZZ
        if self.id == 2:
            return types.Game.GENSHIN
        return self.id


class MimoTask(APIModel):
    """Mimo task."""

    id: int = Aliased("task_id")
    name: str = Aliased("task_name")
    time_type: int
    point: int
    progress: int
    total_progress: int

    status: MimoTaskStatus
    jump_url: str
    window_text: str
    type: typing.Union[int, MimoTaskType] = Aliased("task_type")
    af_url: str

    @pydantic.field_validator("type", mode="before")
    def __transform_task_type(cls, v: int) -> typing.Union[int, MimoTaskType]:
        return prevent_enum_error(v, MimoTaskType)


class MimoShopItem(APIModel):
    """Mimo shop item."""

    id: int = Aliased("award_id")
    status: MimoShopItemStatus
    icon: str
    name: str
    cost: int
    stock: int
    user_count: int
    next_refresh_time: datetime.timedelta
    expire_day: int


class PartialMimoLotteryReward(APIModel):
    """Partial mimo lottery reward."""

    type: int
    icon: str
    name: str


class MimoLotteryReward(PartialMimoLotteryReward):
    """Mimo lottery reward."""

    expire_day: int


class MimoLotteryInfo(APIModel):
    """Mimo lottery info."""

    current_point: int = Aliased("point")
    cost: int
    current_count: int = Aliased("count")
    limit_count: int
    rewards: typing.Sequence[MimoLotteryReward] = Aliased("award_list")


class MimoLotteryResult(APIModel):
    """Mimo lottery result."""

    reward: PartialMimoLotteryReward
    reward_id: int = Aliased("award_id")
    game_id: int
    src_type: int
    code: str = Aliased("exchange_code")

    @pydantic.model_validator(mode="before")
    def __nest_reward(cls, v: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        v["reward"] = {"type": v.pop("type"), "icon": v.pop("icon"), "name": v.pop("name")}
        return v



import typing

from genshin.models.model import Aliased, APIModel

__all__ = ("ChallengeBangboo",)


class ChallengeBangboo(APIModel):
    """Shiyu Defense bangboo model."""

    id: int
    rarity: typing.Literal["S", "A"]
    level: int
    icon: str = Aliased("bangboo_rectangle_url")
